"""Representation Module.

This module contains the functionality related
with plotting and scattering our different datasets.
It allows multiple modes and colors, which could
be set manually or automatically depending on values
like depth measures.
"""

from typing import Any, Dict, Optional, Sequence, Tuple, TypeVar, Union

import matplotlib.cm
import matplotlib.patches
import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from typing_extensions import Protocol

from ... import FDataGrid
from ..._utils import _to_domain_range, _to_grid_points, constants
from ...representation._functional_data import FData
from ...representation._typing import DomainRangeLike, GridPointsLike
from ._baseplot import BasePlot
from ._utils import ColorLike, _set_labels

K = TypeVar('K', contravariant=True)
V = TypeVar('V', covariant=True)
T = TypeVar('T', FDataGrid, np.ndarray)


class Indexable(Protocol[K, V]):
    """Class Indexable used to type _get_color_info."""

    def __getitem__(self, __key: K) -> V:
        pass

    def __len__(self) -> int:
        pass


def _get_color_info(
    fdata: T,
    group: Optional[Sequence[K]] = None,
    group_names: Optional[Indexable[K, str]] = None,
    group_colors: Optional[Indexable[K, ColorLike]] = None,
    legend: bool = False,
    kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[
    Optional[ColorLike],
    Optional[Sequence[matplotlib.patches.Patch]],
]:

    if kwargs is None:
        kwargs = {}

    patches = None

    if group is not None:
        # In this case, each curve has a label, and all curves with the same
        # label should have the same color

        group_unique, group_indexes = np.unique(group, return_inverse=True)
        n_labels = len(group_unique)

        if group_colors is not None:
            group_colors_array = np.array(
                [group_colors[g] for g in group_unique],
            )
        else:
            prop_cycle = matplotlib.rcParams['axes.prop_cycle']
            cycle_colors = prop_cycle.by_key()['color']

            group_colors_array = np.take(
                cycle_colors, np.arange(n_labels), mode='wrap',
            )

        sample_colors = group_colors_array[group_indexes]

        group_names_array = None

        if group_names is not None:
            group_names_array = np.array(
                [group_names[g] for g in group_unique],
            )
        elif legend is True:
            group_names_array = group_unique

        if group_names_array is not None:
            patches = [
                matplotlib.patches.Patch(color=c, label=l)
                for c, l in zip(group_colors_array, group_names_array)
            ]

    else:
        # In this case, each curve has a different color unless specified
        # otherwise

        if 'color' in kwargs:
            sample_colors = len(fdata) * [kwargs.get("color")]
            kwargs.pop('color')

        elif 'c' in kwargs:
            sample_colors = len(fdata) * [kwargs.get("c")]
            kwargs.pop('c')

        else:
            sample_colors = None

    return sample_colors, patches


class GraphPlot(BasePlot):
    """
    Class used to plot the FDataGrid object graph as hypersurfaces.

    When plotting functional data, we can either choose manually a color,
    a group of colors for the representations. Besides, we can use a list of
    variables (depths, scalar regression targets...) can be used as an
    argument to display the functions wtih a gradient of colors.
    Args:
        fdata: functional data set that we want to plot.
        gradient_criteria: list of real values used to determine the color
            in which each of the instances will be plotted.
        max_grad: maximum value that the gradient_list can take, it will be
            used to normalize the ``gradient_criteria`` in order to get values
            that can be used in the function colormap.__call__(). If not
            declared it will be initialized to the maximum value of
            gradient_list.
        min_grad: minimum value that the gradient_list can take, it will be
            used to normalize the ``gradient_criteria`` in order to get values
            that can be used in the function colormap.__call__(). If not
            declared it will be initialized to the minimum value of
            gradient_list.
        chart: figure over
            with the graphs are plotted or axis over where the graphs are
            plotted. If None and ax is also None, the figure is
            initialized.
        fig: figure over with the graphs are
            plotted in case ax is not specified. If None and ax is also
            None, the figure is initialized.
        axes: axis over where the graphs
            are plotted. If None, see param fig.
        n_rows: designates the number of rows of the figure
            to plot the different dimensions of the image. Only specified
            if fig and ax are None.
        n_cols: designates the number of columns of the
            figure to plot the different dimensions of the image. Only
            specified if fig and ax are None.
        n_points: Number of points to evaluate in
            the plot. In case of surfaces a tuple of length 2 can be pased
            with the number of points to plot in each axis, otherwise the
            same number of points will be used in the two axes. By default
            in unidimensional plots will be used 501 points; in surfaces
            will be used 30 points per axis, wich makes a grid with 900
            points.
        domain_range: Range where the
            function will be plotted. In objects with unidimensional domain
            the domain range should be a tuple with the bounds of the
            interval; in the case of surfaces a list with 2 tuples with
            the ranges for each dimension. Default uses the domain range
            of the functional object.
        group: contains integers from [0 to number of
            labels) indicating to which group each sample belongs to. Then,
            the samples with the same label are plotted in the same color.
            If None, the default value, each sample is plotted in the color
            assigned by matplotlib.pyplot.rcParams['axes.prop_cycle'].
        group_colors: colors in which groups are
            represented, there must be one for each group. If None, each
            group is shown with distict colors in the "Greys" colormap.
        group_names: name of each of the groups which appear
            in a legend, there must be one for each one. Defaults to None
            and the legend is not shown. Implies `legend=True`.
        colormap: name of the colormap to be used. By default we will
            use autumn.
        legend: if `True`, show a legend with the groups. If
            `group_names` is passed, it will be used for finding the names
            to display in the legend. Otherwise, the values passed to
            `group` will be used.
        kwargs: if dim_domain is 1, keyword arguments to be passed to
            the matplotlib.pyplot.plot function; if dim_domain is 2,
            keyword arguments to be passed to the
            matplotlib.pyplot.plot_surface function.
    Attributes:
        gradient_list: normalization of the values from gradient color_list
            that will be used to determine the intensity of the color
            each function will have.
    """

    def __init__(
        self,
        fdata: FData,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Optional[Axes] = None,
        n_rows: Optional[int] = None,
        n_cols: Optional[int] = None,
        n_points: Union[int, Tuple[int, int], None] = None,
        domain_range: Optional[DomainRangeLike] = None,
        group: Optional[Sequence[K]] = None,
        group_colors: Optional[Indexable[K, ColorLike]] = None,
        group_names: Optional[Indexable[K, str]] = None,
        gradient_criteria: Optional[Sequence[float]] = None,
        max_grad: Optional[float] = None,
        min_grad: Optional[float] = None,
        colormap: Union[Colormap, str, None] = None,
        legend: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            chart,
            fig=fig,
            axes=axes,
            n_rows=n_rows,
            n_cols=n_cols,
        )
        self.fdata = fdata
        self.gradient_criteria = gradient_criteria
        if self.gradient_criteria is not None:
            if len(self.gradient_criteria) != fdata.n_samples:
                raise ValueError(
                    "The length of the gradient color",
                    "list should be the same as the number",
                    "of samples in fdata",
                )

            if min_grad is None:
                self.min_grad = min(self.gradient_criteria)
            else:
                self.min_grad = min_grad

            if max_grad is None:
                self.max_grad = max(self.gradient_criteria)
            else:
                self.max_grad = max_grad

            self.gradient_list: Optional[Sequence[float]] = (
                [
                    (grad_color - self.min_grad)
                    / (self.max_grad - self.min_grad)
                    for grad_color in self.gradient_criteria
                ]
            )
        else:
            self.gradient_list = None

        self.n_points = n_points
        self.group = group
        self.group_colors = group_colors
        self.group_names = group_names
        self.legend = legend
        self.colormap = colormap
        self.kwargs = kwargs

        if domain_range is None:
            self.domain_range = self.fdata.domain_range
        else:
            self.domain_range = _to_domain_range(domain_range)

        if self.gradient_list is None:
            sample_colors, patches = _get_color_info(
                self.fdata,
                self.group,
                self.group_names,
                self.group_colors,
                self.legend,
                kwargs,
            )
        else:
            patches = None
            if self.colormap is None:
                colormap = matplotlib.cm.get_cmap("autumn")
                colormap = colormap.reversed()
            else:
                colormap = matplotlib.cm.get_cmap(self.colormap)

            sample_colors = colormap(self.gradient_list)

        self.sample_colors = sample_colors
        self.patches = patches

    @property
    def dim(self) -> int:
        return self.fdata.dim_domain + 1

    @property
    def n_subplots(self) -> int:
        return self.fdata.dim_codomain

    @property
    def n_samples(self) -> int:
        return self.fdata.n_samples

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:

        self.artists = np.zeros(
            (self.n_samples, self.fdata.dim_codomain),
            dtype=Artist,
        )

        color_dict: Dict[str, Optional[ColorLike]] = {}

        if self.fdata.dim_domain == 1:

            if self.n_points is None:
                self.n_points = constants.N_POINTS_UNIDIMENSIONAL_PLOT_MESH

            assert isinstance(self.n_points, int)

            # Evaluates the object in a linspace
            eval_points = np.linspace(*self.domain_range[0], self.n_points)
            mat = self.fdata(eval_points)

            for i in range(self.fdata.dim_codomain):
                for j in range(self.fdata.n_samples):

                    set_color_dict(self.sample_colors, j, color_dict)

                    self.artists[j, i] = axes[i].plot(
                        eval_points,
                        mat[j, ..., i].T,
                        **self.kwargs,
                        **color_dict,
                    )[0]

        else:

            # Selects the number of points
            if self.n_points is None:
                n_points_tuple = 2 * (constants.N_POINTS_SURFACE_PLOT_AX,)
            elif isinstance(self.n_points, int):
                n_points_tuple = (self.n_points, self.n_points)
            elif len(self.n_points) != 2:
                raise ValueError(
                    "n_points should be a number or a tuple of "
                    "length 2, and has "
                    "length {0}.".format(len(self.n_points)),
                )

            # Axes where will be evaluated
            x = np.linspace(*self.domain_range[0], n_points_tuple[0])
            y = np.linspace(*self.domain_range[1], n_points_tuple[1])

            # Evaluation of the functional object
            Z = self.fdata((x, y), grid=True)

            X, Y = np.meshgrid(x, y, indexing='ij')

            for k in range(self.fdata.dim_codomain):
                for h in range(self.fdata.n_samples):

                    set_color_dict(self.sample_colors, h, color_dict)

                    self.artists[h, k] = axes[k].plot_surface(
                        X,
                        Y,
                        Z[h, ..., k],
                        **self.kwargs,
                        **color_dict,
                    )

        _set_labels(self.fdata, fig, axes, self.patches)


class ScatterPlot(BasePlot):
    """
    Class used to scatter the FDataGrid object.

    Args:
        fdata: functional data set that we want to plot.
        grid_points: points to plot.
        chart: figure over
            with the graphs are plotted or axis over where the graphs are
            plotted. If None and ax is also None, the figure is
            initialized.
        fig: figure over with the graphs are
            plotted in case ax is not specified. If None and ax is also
            None, the figure is initialized.
        axes: axis over where the graphs
            are plotted. If None, see param fig.
        n_rows: designates the number of rows of the figure
            to plot the different dimensions of the image. Only specified
            if fig and ax are None.
        n_cols: designates the number of columns of the
            figure to plot the different dimensions of the image. Only
            specified if fig and ax are None.
        domain_range: Range where the
            function will be plotted. In objects with unidimensional domain
            the domain range should be a tuple with the bounds of the
            interval; in the case of surfaces a list with 2 tuples with
            the ranges for each dimension. Default uses the domain range
            of the functional object.
        group: contains integers from [0 to number of
            labels) indicating to which group each sample belongs to. Then,
            the samples with the same label are plotted in the same color.
            If None, the default value, each sample is plotted in the color
            assigned by matplotlib.pyplot.rcParams['axes.prop_cycle'].
        group_colors: colors in which groups are
            represented, there must be one for each group. If None, each
            group is shown with distict colors in the "Greys" colormap.
        group_names: name of each of the groups which appear
            in a legend, there must be one for each one. Defaults to None
            and the legend is not shown. Implies `legend=True`.
        legend: if `True`, show a legend with the groups. If
            `group_names` is passed, it will be used for finding the names
            to display in the legend. Otherwise, the values passed to
            `group` will be used.
        kwargs: if dim_domain is 1, keyword arguments to be passed to
            the matplotlib.pyplot.plot function; if dim_domain is 2,
            keyword arguments to be passed to the
            matplotlib.pyplot.plot_surface function.
    """

    def __init__(
        self,
        fdata: FData,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Optional[Axes] = None,
        n_rows: Optional[int] = None,
        n_cols: Optional[int] = None,
        grid_points: Optional[GridPointsLike] = None,
        domain_range: Union[Tuple[int, int], DomainRangeLike, None] = None,
        group: Optional[Sequence[K]] = None,
        group_colors: Optional[Indexable[K, ColorLike]] = None,
        group_names: Optional[Indexable[K, str]] = None,
        legend: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            chart,
            fig=fig,
            axes=axes,
            n_rows=n_rows,
            n_cols=n_cols,
        )
        self.fdata = fdata

        if grid_points is None:
            # This can only be done for FDataGrid
            self.grid_points = self.fdata.grid_points
            self.evaluated_points = self.fdata.data_matrix
        else:
            self.grid_points = _to_grid_points(grid_points)
            self.evaluated_points = self.fdata(
                self.grid_points, grid=True,
            )

        self.domain_range = domain_range
        self.group = group
        self.group_colors = group_colors
        self.group_names = group_names
        self.legend = legend

        if self.domain_range is None:
            self.domain_range = self.fdata.domain_range
        else:
            self.domain_range = _to_domain_range(self.domain_range)

        sample_colors, patches = _get_color_info(
            self.fdata,
            self.group,
            self.group_names,
            self.group_colors,
            self.legend,
            kwargs,
        )
        self.sample_colors = sample_colors
        self.patches = patches

    @property
    def dim(self) -> int:
        return self.fdata.dim_domain + 1

    @property
    def n_subplots(self) -> int:
        return self.fdata.dim_codomain

    @property
    def n_samples(self) -> int:
        return self.fdata.n_samples

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:
        """
        Scatter FDataGrid object.

        Returns:
        fig: figure object in which the graphs are plotted.
        """
        self.artists = np.zeros(
            (self.n_samples, self.fdata.dim_codomain),
            dtype=Artist,
        )

        color_dict: Dict[str, Optional[ColorLike]] = {}

        if self.fdata.dim_domain == 1:

            for i in range(self.fdata.dim_codomain):
                for j in range(self.fdata.n_samples):

                    set_color_dict(self.sample_colors, j, color_dict)

                    self.artists[j, i] = axes[i].scatter(
                        self.grid_points[0],
                        self.evaluated_points[j, ..., i].T,
                        **color_dict,
                        picker=True,
                        pickradius=2,
                    )

        else:

            X = self.fdata.grid_points[0]
            Y = self.fdata.grid_points[1]
            X, Y = np.meshgrid(X, Y)

            for k in range(self.fdata.dim_codomain):
                for h in range(self.fdata.n_samples):

                    set_color_dict(self.sample_colors, h, color_dict)

                    self.artists[h, k] = axes[k].scatter(
                        X,
                        Y,
                        self.evaluated_points[h, ..., k].T,
                        **color_dict,
                        picker=True,
                        pickradius=2,
                    )

        _set_labels(self.fdata, fig, axes, self.patches)


def set_color_dict(
    sample_colors: Any,
    ind: int,
    color_dict: Dict[str, Optional[ColorLike]],
) -> None:
    """
    Auxiliary method used to update color_dict.

    Sets the new color of the color_dict
    thanks to sample colors and index.
    """
    if sample_colors is not None:
        color_dict["color"] = sample_colors[ind]



"""DD-Plot Module.

This module contains the necessary functions to construct the DD-Plot.
To do this depth is calculated for the two chosen distributions, and then
a scatter plot is created of this two variables.
"""

from typing import Optional, TypeVar, Union

import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure

from ...exploratory.depth.multivariate import Depth
from ...representation._functional_data import FData
from ...representation._typing import NDArrayInt
from ._baseplot import BasePlot

T = TypeVar('T', bound=FData)


class DDPlot(BasePlot):
    """
    DDPlot visualization.

    Plot the depth of our fdata elements in two
    different distributions, one in each axis. It is useful to understand
    how our data is more related with one subset of data / distribution
    than another one.
    Args:
        fdata: functional data set that we want to examine.
        dist1: functional data set that represents the first distribution that
            we want to use to compute the depth (Depth X).
        dist2: functional data set that represents the second distribution that
            we want to use to compute the depth (Depth Y).
        depth_method: method that will be used to compute the depths of the
            data with respect to the distributions.
        chart: figure over with the graphs are plotted or axis over
            where the graphs are plotted. If None and ax is also
            None, the figure is initialized.
        fig: figure over with the graphs are plotted in case ax is not
            specified. If None and ax is also None, the figure is
            initialized.
        axes: axis where the graphs are plotted. If None, see param fig.
    Attributes:
        depth_dist1: result of the calculation of the depth_method into our
            first distribution (dist1).
        depth_dist2: result of the calculation of the depth_method into our
            second distribution (dist2).
    """

    def __init__(
        self,
        fdata: T,
        dist1: T,
        dist2: T,
        chart: Union[Figure, Axes, None] = None,
        *,
        depth_method: Depth[T],
        fig: Optional[Figure] = None,
        axes: Optional[Axes] = None,
        c: Optional[NDArrayInt] = None,
        cmap_bold: ListedColormap = None,
        x_label: str = "X depth",
        y_label: str = "Y depth",
    ) -> None:
        super().__init__(
            chart,
            fig=fig,
            axes=axes,
            c=c,
            cmap_bold=cmap_bold,
            x_label=x_label,
            y_label=y_label,
        )
        self.fdata = fdata
        self.depth_method = depth_method
        self.depth_dist1 = self.depth_method(
            self.fdata,
            distribution=dist1,
        )
        self.depth_dist2 = self.depth_method(
            self.fdata,
            distribution=dist2,
        )

    @property
    def n_samples(self) -> int:
        return self.fdata.n_samples

    def _plot(
        self,
        fig: Figure,
        axes: Axes,
    ) -> None:
        """
        Plot DDPlot graph.

        Plot the depth of our fdata elements in the two different
        distributions, one in each axis. It is useful to understand how
        our data is more related with one subset of data / distribution
        than another one.
        Returns:
            fig (figure object): figure object in which the depths will be
            scattered.
        """
        self.artists = np.zeros(
            (self.n_samples, 1),
            dtype=Artist,
        )
        margin = 0.025
        width_aux_line = 1
        color_aux_line = "gray"

        ax = axes[0]

        self.artists[:, 0] = ax.scatter(
            self.depth_dist1,
            self.depth_dist2,
            c=self.c,
            cmap=self.cmap_bold,
            picker=True,
            pickradius=2,
        )

        # Set labels of graph
        if self.fdata.dataset_name is not None:
            ax.set_title(self.fdata.dataset_name)
        ax.set_xlim(
            [
                self.depth_method.min - margin,
                self.depth_method.max + margin,
            ],
        )
        ax.set_ylim(
            [
                self.depth_method.min - margin,
                self.depth_method.max + margin,
            ],
        )
        ax.plot(
            [0, 1],
            linewidth=width_aux_line,
            color=color_aux_line,
        )



"""Test to check the per class transformer module."""

import unittest

import numpy as np

from skfda._utils import _classifier_get_classes
from skfda.datasets import fetch_growth
from skfda.ml.classification import KNeighborsClassifier
from skfda.preprocessing.dim_reduction.feature_extraction import (
    PerClassTransformer,
)
from skfda.preprocessing.dim_reduction.variable_selection import (
    RecursiveMaximaHunting,
)


class TestPerClassTransformer(unittest.TestCase):
    """Tests for PCT module."""

    def setUp(self) -> None:
        """Fetch the Berkeley Growth Study dataset."""
        X, y = fetch_growth(return_X_y=True, as_frame=True)
        self.X = X.iloc[:, 0].values
        self.y = y.values.codes

    def test_transform(self) -> None:
        """Check the data transformation is done correctly."""
        t = PerClassTransformer(
            RecursiveMaximaHunting(),  # type: ignore
            array_output=True,
        )
        t.fit_transform(self.X, self.y)
        transformed = t.transform(self.X)

        manual = np.empty((93, 0))
        classes, y_ind = _classifier_get_classes(self.y)
        for cur_class in range(classes.size):
            feature_transformer = RecursiveMaximaHunting().fit(  # type: ignore
                self.X[y_ind == cur_class],
                self.y[y_ind == cur_class],
            )
            aux = np.array(feature_transformer.transform(self.X))
            manual = np.hstack((manual, aux))

        np.testing.assert_array_equal(transformed, manual)

    def test_not_transformer_argument(self) -> None:
        """Check that invalid arguments in fit raise exception."""
        t = PerClassTransformer(KNeighborsClassifier())  # type: ignore
        self.assertRaises(
            TypeError,
            t.fit,
            self.X,
            self.y,
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import Tuple

import numpy as np
import numpy.linalg as linalg
import sklearn.utils.validation

from ...._utils import _classifier_get_classes
from ....representation import FDataGrid


def _rkhs_vs(
    X: np.ndarray,
    Y: np.ndarray,
    n_features_to_select: int = 1,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    RKHS-VS implementation.

    Parameters:
        X: Matrix of trajectories
        Y: Vector of class labels
        n_features_to_select: Number of selected features

    Returns:
        Selected features and vector of scores.

    """
    X = np.atleast_2d(X)
    assert n_features_to_select >= 1
    assert n_features_to_select <= X.shape[1]

    _, Y = _classifier_get_classes(Y)

    selected_features = np.zeros(n_features_to_select, dtype=int)
    score = np.zeros(n_features_to_select)
    indexes = np.arange(0, X.shape[1])

    # Calculate means and covariance matrix
    class_1_trajectories = X[Y.ravel() == 1]
    class_0_trajectories = X[Y.ravel() == 0]

    means = (
        np.mean(class_1_trajectories, axis=0)
        - np.mean(class_0_trajectories, axis=0)
    )

    class_1_count = sum(Y)
    class_0_count = Y.shape[0] - class_1_count

    class_1_proportion = class_1_count / Y.shape[0]
    class_0_proportion = class_0_count / Y.shape[0]

    # The result should be casted to 2D because of bug #11502 in numpy
    variances = (
        class_1_proportion * np.atleast_2d(
            np.cov(class_1_trajectories, rowvar=False, bias=True),
        )
        + class_0_proportion * np.atleast_2d(
            np.cov(class_0_trajectories, rowvar=False, bias=True),
        )
    )

    # The first variable maximizes |mu(t)|/sigma(t)
    mu_sigma = np.abs(means) / np.sqrt(np.diag(variances))

    selected_features[0] = np.argmax(mu_sigma)
    score[0] = mu_sigma[selected_features[0]]
    indexes = np.delete(indexes, selected_features[0])

    for i in range(1, n_features_to_select):
        aux = np.zeros_like(indexes, dtype=np.float_)

        for j in range(0, indexes.shape[0]):
            new_selection = np.concatenate([
                selected_features[:i],
                [indexes[j]],
            ])

            new_means = np.atleast_2d(means[new_selection])

            lstsq_solution = linalg.lstsq(
                variances[new_selection[:, np.newaxis], new_selection],
                new_means.T,
                rcond=None,
            )[0]

            aux[j] = new_means @ lstsq_solution

        aux2 = np.argmax(aux)
        selected_features[i] = indexes[aux2]
        score[i] = aux[aux2]
        indexes = np.delete(indexes, aux2)

    return selected_features, score


class RKHSVariableSelection(
    sklearn.base.BaseEstimator,  # type: ignore
    sklearn.base.TransformerMixin,  # type: ignore
):
    r"""
    Reproducing kernel variable selection.

    This is a filter variable selection method for binary classification
    problems. With a fixed number :math:`d` of variables to select, it aims to
    find the variables :math:`X(t_1), \ldots, X(t_d)` for the values
    :math:`t_1, \ldots, t_d` that maximize the separation of
    the class means in the reduced space, measured using the Mahalanobis
    distance

    .. math::
        \phi(t_1, \ldots, t_d) = m_{t_1, \ldots, t_d}^T
        K_{t_1, \ldots, t_d}^{-1} m_{t_1, \ldots, t_d}

    where :math:`m_{t_1, \ldots, t_d}` is the difference of the mean
    functions of both classes evaluated at points :math:`t_1, \ldots, t_d`
    and :math:`K_{t_1, \ldots, t_d}` is the common covariance function
    evaluated at the same points.

    This method is optimal, with a fixed value of :math:`d`, for variable
    selection in Gaussian binary classification problems with the same
    covariance in both classes (homoscedasticity), when all possible
    combinations of points are taken into account. That means that for all
    possible selections of :math:`t_1, \ldots, t_d`, the one in which
    :math:`\phi(t_1, \ldots, t_d)` is greater minimizes the optimal
    misclassification error of all the classification problems with the
    reduced dimensionality. For a longer discussion about the optimality and
    consistence of this method, we refer the reader to the original
    article [1]_.

    In practice the points are selected one at a time, using
    a greedy approach, so this optimality is not always guaranteed.

    Parameters:
        n_features_to_select: number of features to select.

    Examples:
        >>> from skfda.preprocessing.dim_reduction import variable_selection
        >>> from skfda.datasets import make_gaussian_process
        >>> import skfda
        >>> import numpy as np

        We create trajectories from two classes, one with zero mean and the
        other with a peak-like mean. Both have Brownian covariance.

        >>> n_samples = 10000
        >>> n_features = 1000
        >>>
        >>> def mean_1(t):
        ...     return (np.abs(t - 0.25)
        ...             - 2 * np.abs(t - 0.5)
        ...             + np.abs(t - 0.75))
        >>>
        >>> X_0 = make_gaussian_process(n_samples=n_samples // 2,
        ...                             n_features=n_features,
        ...                             random_state=0)
        >>> X_1 = make_gaussian_process(n_samples=n_samples // 2,
        ...                             n_features=n_features,
        ...                             mean=mean_1,
        ...                             random_state=1)
        >>> X = skfda.concatenate((X_0, X_1))
        >>>
        >>> y = np.zeros(n_samples)
        >>> y [n_samples // 2:] = 1

        Select the relevant points to distinguish the two classes

        >>> rkvs = variable_selection.RKHSVariableSelection(
        ...                               n_features_to_select=3)
        >>> _ = rkvs.fit(X, y)
        >>> point_mask = rkvs.get_support()
        >>> points = X.grid_points[0][point_mask]
        >>> np.allclose(points, [0.25, 0.5, 0.75], rtol=1e-2)
        True

        Apply the learned dimensionality reduction

        >>> X_dimred = rkvs.transform(X)
        >>> len(X.grid_points[0])
        1000
        >>> X_dimred.shape
        (10000, 3)

    References:
        .. [1] J. R. Berrendero, A. Cuevas, and J. L. Torrecilla, «On the Use
               of Reproducing Kernel Hilbert Spaces in Functional
               Classification», Journal of the American Statistical
               Association, vol. 113, no. 523, pp. 1210-1218, jul. 2018,
               doi: 10.1080/01621459.2017.1320287.

    """

    def __init__(self, n_features_to_select: int = 1) -> None:
        self.n_features_to_select = n_features_to_select

    def fit(  # noqa: D102
        self,
        X: FDataGrid,
        y: np.ndarray,
    ) -> RKHSVariableSelection:

        n_unique_labels = len(np.unique(y))
        if n_unique_labels != 2:
            raise ValueError(
                f"RK-VS can only be used when there are only "
                f"two different labels, but there are "
                f"{n_unique_labels}",
            )

        if X.dim_domain != 1 or X.dim_codomain != 1:
            raise ValueError("Domain and codomain dimensions must be 1")

        X, y = sklearn.utils.validation.check_X_y(X.data_matrix[..., 0], y)

        self._features_shape_ = X.shape[1:]

        features, scores = _rkhs_vs(
            X=X,
            Y=y,
            n_features_to_select=self.n_features_to_select,
        )

        self._features_ = features
        self._scores_ = scores

        return self

    def transform(  # noqa: D102
        self,
        X: FDataGrid,
        Y: None = None,
    ) -> np.ndarray:

        sklearn.utils.validation.check_is_fitted(self)

        X_matrix = sklearn.utils.validation.check_array(X.data_matrix[..., 0])

        if X_matrix.shape[1:] != self._features_shape_:
            raise ValueError(
                "The trajectories have a different number of "
                "points than the ones fitted",
            )

        return X_matrix[:, self._features_]

    def get_support(self, indices: bool = False) -> np.ndarray:
        """
        Get a mask, or integer index, of the features selected.

        Parameters:
            indices: If True, the return value will be an array of integers,
                rather than a boolean mask.

        Returns:
            An index that selects the retained features from a `FDataGrid`
            object.
            If `indices` is False, this is a boolean array of shape
            [# input features], in which an element is True iff its
            corresponding feature is selected for retention. If `indices`
            is True, this is an integer array of shape [# output features]
            whose values are indices into the input feature vector.

        """
        features = self._features_
        if indices:
            return features

        mask = np.zeros(self._features_shape_[0], dtype=bool)
        mask[features] = True
        return mask



"""Parametric Plot Module.

This module contains the functionality in charge of plotting
two different functions as coordinates, this can be done giving
one FData, with domain 1 and codomain 2, or giving two FData, both
of them with domain 1 and codomain 1.
"""

from typing import Mapping, Optional, Sequence, TypeVar, Union

import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ...representation import FData
from ._baseplot import BasePlot
from ._utils import ColorLike
from .representation import Indexable, _get_color_info

K = TypeVar('K', contravariant=True)


class ParametricPlot(BasePlot):
    """
    Parametric Plot visualization.

    This class contains the functionality in charge of plotting
    two different functions as coordinates, this can be done giving
    one FData, with domain 1 and codomain 2, or giving two FData, both
    of them with domain 1 and codomain 1.
    Args:
        fdata1: functional data set that we will use for the graph. If it has
            a dim_codomain = 1, the fdata2 will be needed.
        fdata2: optional functional data set, that will be needed if the fdata1
            has dim_codomain = 1.
        chart: figure over with the graphs are plotted or axis over
            where the graphs are plotted. If None and ax is also
            None, the figure is initialized.
        fig: figure over with the graphs are plotted in case ax is not
            specified. If None and ax is also None, the figure is
            initialized.
        ax: axis where the graphs are plotted. If None, see param fig.
    """

    def __init__(
        self,
        fdata1: FData,
        fdata2: Optional[FData] = None,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Optional[Axes] = None,
        group: Optional[Sequence[K]] = None,
        group_colors: Optional[Indexable[K, ColorLike]] = None,
        group_names: Optional[Indexable[K, str]] = None,
        legend: bool = False,
    ) -> None:
        BasePlot.__init__(
            self,
            chart,
            fig=fig,
            axes=axes,
        )
        self.fdata1 = fdata1
        self.fdata2 = fdata2

        if self.fdata2 is not None:
            self.fd_final = self.fdata1.concatenate(
                self.fdata2, as_coordinates=True,
            )
        else:
            self.fd_final = self.fdata1

        self.group = group
        self.group_names = group_names
        self.group_colors = group_colors
        self.legend = legend

    @property
    def n_samples(self) -> int:
        return self.fd_final.n_samples

    def _plot(
        self,
        fig: Figure,
        axes: Axes,
    ) -> None:

        self.artists = np.zeros((self.n_samples, 1), dtype=Artist)

        sample_colors, patches = _get_color_info(
            self.fd_final,
            self.group,
            self.group_names,
            self.group_colors,
            self.legend,
        )

        color_dict: Mapping[str, Union[ColorLike, None]] = {}

        if (
            self.fd_final.dim_domain == 1
            and self.fd_final.dim_codomain == 2
        ):
            ax = axes[0]

            for i in range(self.fd_final.n_samples):

                if sample_colors is not None:
                    color_dict["color"] = sample_colors[i]

                self.artists[i, 0] = ax.plot(
                    self.fd_final.data_matrix[i][:, 0].tolist(),
                    self.fd_final.data_matrix[i][:, 1].tolist(),
                    **color_dict,
                )[0]

        else:
            raise ValueError(
                "Error in data arguments,",
                "codomain or domain is not correct.",
            )

        if self.fd_final.dataset_name is not None:
            fig.suptitle(self.fd_final.dataset_name)

        if self.fd_final.coordinate_names[0] is None:
            ax.set_xlabel("Function 1")
        else:
            ax.set_xlabel(self.fd_final.coordinate_names[0])

        if self.fd_final.coordinate_names[1] is None:
            ax.set_ylabel("Function 2")
        else:
            ax.set_ylabel(self.fd_final.coordinate_names[1])



"""Feature extraction transformers for dimensionality reduction."""
from __future__ import annotations

import warnings
from typing import Any, Mapping, TypeVar, Union

import numpy as np
from pandas import DataFrame
from sklearn.utils.validation import check_is_fitted as sklearn_check_is_fitted

from ...._utils import TransformerMixin, _fit_feature_transformer
from ....representation import FData
from ....representation._typing import NDArrayFloat, NDArrayInt
from ....representation.basis import FDataBasis
from ....representation.grid import FDataGrid

Input = TypeVar("Input", bound=Union[FData, NDArrayFloat])
Output = TypeVar("Output", bound=Union[DataFrame, NDArrayFloat])
Target = TypeVar("Target", bound=NDArrayInt)

TransformerOutput = Union[FData, NDArrayFloat]


class PerClassTransformer(TransformerMixin[Input, Output, Target]):
    r"""Per class feature transformer for functional data.

    This class takes a transformer and performs the following map:

    .. math::
        \mathcal{X} &\rightarrow \mathbb{R}^G \\
        x &\rightarrow \textbf{t} = (T_1(x), T_2(x),...,T_k(x))

    Where :math:`T_i(x)` is the transformation  :math:`x` with respect to
    the data in the :math:`i`-th group.

    Note that :math:`\mathcal{X}` is possibly multivariate, that is,
    :math:`\mathcal{X} = \mathcal{X}_1 \times ... \times \mathcal{X}_p`.

    Parameters:
        transformer:
            The transformer that we want to apply to the given data.
            It should use target data while fitting.
            This is checked by looking at the 'stateless' and 'requires_y' tags
        array_output:
            Indicates if the transformed data is requested to be a NumPy array
            output. By default the value is False.

    Examples:
        Firstly, we will import the Berkeley Growth Study dataset:

        >>> from skfda.datasets import fetch_growth
        >>> X, y = fetch_growth(return_X_y=True, as_frame=True)
        >>> X = X.iloc[:, 0].values
        >>> y = y.values.codes

        >>> from skfda.preprocessing.dim_reduction.feature_extraction import (
        ...     PerClassTransformer,
        ... )

        Then we will need to select a fda transformer, and so we will
        use RecursiveMaximaHunting. We need to fit the data and transform it:

        >>> from skfda.preprocessing.dim_reduction.variable_selection import (
        ...     RecursiveMaximaHunting,
        ... )
        >>> t1 = PerClassTransformer(
        ...     RecursiveMaximaHunting(),
        ...     array_output=True,
        ... )
        >>> x_transformed1 = t1.fit_transform(X, y)

        ``x_transformed1`` will be a vector with the transformed data.
        We will split the generated data and fit a KNN classifier.

        >>> from sklearn.model_selection import train_test_split
        >>> from sklearn.neighbors import KNeighborsClassifier
        >>> X_train1, X_test1, y_train1, y_test1 = train_test_split(
        ...     x_transformed1,
        ...     y,
        ...     test_size=0.25,
        ...     stratify=y,
        ...     random_state=0,
        ... )
        >>> neigh1 = KNeighborsClassifier()
        >>> neigh1 = neigh1.fit(X_train1, y_train1)

        Finally we can predict and check the score:
        >>> neigh1.predict(X_test1)
            array([0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
            1, 1, 1, 1, 1, 1, 1], dtype=int8)

        >>> round(neigh1.score(X_test1, y_test1), 3)
            0.958



        We can also use a transformer that returns a FData object
        when predicting.
        In our example we are going to use the Nadaraya Watson Smoother.

        >>> from skfda.preprocessing.smoothing.kernel_smoothers import (
        ...     NadarayaWatsonSmoother,
        ... )
        >>> t2 = PerClassTransformer(
        ...     NadarayaWatsonSmoother(),
        ... )
        >>> x_transformed2 = t2.fit_transform(X, y)

        ``x_transformed2`` will be a DataFrame with the transformed data.
        Each row on the frame contains a FDataGrid describing a transformed
        curve.
        We need to convert the DataFrame into a FDataGrid with all the
        samples, so we can train a classifier. We also need to duplicate
        the outputs as we have the double amount of data curves:

        >>> for i, curve_grid in enumerate(x_transformed2.iloc[:,0].values):
        ...     if i == 0:
        ...         X_transformed_grid = curve_grid
        ...     else:
        ...         X_transformed_grid = X_transformed_grid.concatenate(
        ...                                 curve_grid,
        ...                              )

        >>> y = np.concatenate((y,y))


        ``X_transformed_grid`` contains a FDataGrid with all the transformed
        curves. Now we are able to use it to fit a KNN classifier.
        Again we split the data into train and test.
        >>> X_train2, X_test2, y_train2, y_test2 = train_test_split(
        ...     X_transformed_grid,
        ...     y,
        ...     test_size=0.25,
        ...     stratify=y,
        ...     random_state=0,
        ... )

        This time we need a functional data classifier.
        We fit the classifier and predict.
        >>> from skfda.ml.classification import KNeighborsClassifier
        >>> neigh2 = KNeighborsClassifier()
        >>> neigh2 = neigh2.fit(X_train2, y_train2)
        >>> neigh2.predict(X_test2)
            array([1, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1,
            0, 1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0,
            1, 1, 1, 0, 0], dtype=int8)

        >>> round(neigh2.score(X_test2, y_test2), 3)
            0.957

    """

    def __init__(
        self,
        transformer: TransformerMixin[Input, TransformerOutput, Target],
        *,
        array_output: bool = False,
    ) -> None:
        self.transformer = transformer
        self.array_output = array_output

    def _more_tags(self) -> Mapping[str, Any]:
        parent_tags = super()._more_tags()
        transformer_tags = self.transformer._get_tags()  # noqa: WPS437

        return {
            **parent_tags,
            'allow_nan': transformer_tags['allow_nan'],
            'non_deterministic': transformer_tags['non_deterministic'],
            'pairwise': transformer_tags['pairwise'],
            'requires_positive_X': transformer_tags['requires_positive_X'],
            'requires_y': True,
            'X_types': transformer_tags['X_types'],
        }

    def _validate_transformer(
        self,
    ) -> None:
        """
        Check that the transformer passed is valid.

        Check that it is scikit-learn-like and that
        uses target data in fit.

        Args:
            None

        Returns:
            None
        """
        if not (
            hasattr(self.transformer, "fit")  # noqa: WPS421
            and hasattr(self.transformer, "transform")  # noqa: WPS421
            and hasattr(self.transformer, "fit_transform")  # noqa: WPS421
        ):
            raise TypeError(
                "Transformer should implement fit and "
                "transform. " + str(self.transformer)
                + " (type " + str(type(self.transformer)) + ")"
                " doesn't",
            )

        tags = self.transformer._get_tags()  # noqa: WPS437

        if tags['stateless']:
            warnings.warn(
                f"Parameter 'transformer' with type "
                f"{type(self.transformer)} should use the data for "
                f" fitting."
                f"It should have the 'stateless' tag set to 'False'",
            )

        if tags['requires_y']:
            warnings.warn(
                f"Parameter 'transformer' with type "  # noqa: WPS237
                f"{type(self.transformer)} should not use the class label."
                f"It should have the 'requires_y' tag set to 'False'",
            )

    def fit(  # type: ignore[override]
        self,
        X: Input,
        y: Target,
    ) -> PerClassTransformer[Input, Output, Target]:
        """
        Fit the model on each class.

        It uses X as training data and y as target values.

        Args:
            X: FDataGrid with the training data.
            y: Target values of shape = (n_samples).

        Returns:
            self
        """
        self._validate_transformer()
        classes, class_feature_transformers = _fit_feature_transformer(
            X,
            y,
            self.transformer,
        )

        self._classes = classes
        self._class_feature_transformers_ = class_feature_transformers

        return self

    def transform(self, X: Input) -> Output:
        """
        Transform the provided data using the already fitted transformer.

        Args:
            X: FDataGrid with the test samples.

        Returns:
            Eiter array of shape (n_samples, G) or a Data Frame
            including the transformed data.
        """
        sklearn_check_is_fitted(self)

        transformed_data = [
            feature_transformer.transform(X)
            for feature_transformer in self._class_feature_transformers_
        ]

        if self.array_output:
            for data in transformed_data:
                if isinstance(data, (FDataGrid, FDataBasis)):
                    raise TypeError(
                        "There are transformed instances of FDataGrid or "
                        "FDataBasis that can't be concatenated on a NumPy "
                        "array.",
                    )
            return np.hstack(transformed_data)

        return DataFrame(
            {'Transformed data': transformed_data},
        )

    def fit_transform(  # type: ignore[override]
        self,
        X: Input,
        y: Target,
    ) -> Output:
        """
        Fits and transforms the provided data.

        It uses the transformer specified when initializing the class.

        Args:
            X: FDataGrid with the samples.
            y: Target values of shape = (n_samples)

        Returns:
            Eiter array of shape (n_samples, G) or a Data Frame \
            including the transformed data.
        """
        return self.fit(X, y).transform(X)



import skfda
from skfda.exploratory.depth import IntegratedDepth, ModifiedBandDepth
import unittest
import numpy as np


class TestsDepthSameCurves(unittest.TestCase):

    def setUp(self):
        data_matrix = [[1, 2, 3, 4],
                       [1, 2, 3, 4],
                       [1, 2, 3, 4],
                       [1, 2, 3, 4],
                       [1, 2, 3, 4]]

        self.fd = skfda.FDataGrid(data_matrix)

    def test_integrated_equal(self):

        depth = IntegratedDepth()

        np.testing.assert_almost_equal(
            depth(self.fd), [0.5, 0.5, 0.5, 0.5, 0.5])

    def test_modified_band_depth_equal(self):

        depth = ModifiedBandDepth()

        np.testing.assert_almost_equal(
            depth(self.fd), [1, 1, 1, 1, 1])



"""Neighbors outlier detection methods."""
from sklearn.base import OutlierMixin

from ...misc.metrics import l2_distance
from ...ml._neighbors_base import (
    KNeighborsMixin,
    NeighborsBase,
    NeighborsMixin,
    _to_multivariate_metric,
)


class LocalOutlierFactor(NeighborsBase, NeighborsMixin, KNeighborsMixin,
                         OutlierMixin):
    """Unsupervised Outlier Detection.

    Unsupervised Outlier Detection using Local Outlier Factor (LOF).

    The anomaly score of each sample is called Local Outlier Factor.
    It measures the local deviation of density of a given sample with
    respect to its neighbors.

    It is local in that the anomaly score depends on how isolated the object
    is with respect to the surrounding neighborhood.

    More precisely, locality is given by k-nearest neighbors, whose distance
    is used to estimate the local density.

    By comparing the local density of a sample to the local densities of
    its neighbors, one can identify samples that have a substantially lower
    density than their neighbors. These are considered outliers.

    Parameters
    ----------
    n_neighbors : int, optional (default=20)
        Number of neighbors to use by default for :meth:`kneighbors` queries.
        If n_neighbors is larger than the number of samples provided,
        all samples will be used.
    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use :class:`BallTree`
        - 'kd_tree' will use :class:`KDTree`
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm
          based on the values passed to :meth:`fit` method.

    leaf_size : int, optional (default=30)
        Leaf size passed to :class:`BallTree` or :class:`KDTree`. This can
        affect the speed of the construction and query, as well as the memory
        required to store the tree. The optimal value depends on the
        nature of the problem.
    metric : string or callable, (default
        :func:`l2_distance <skfda.misc.metrics.l2_distance>`)
        the distance metric to use for the tree.  The default metric is
        the L2 distance. See the documentation of the metrics module
        for a list of available metrics.
    metric_params : dict, optional (default=None)
        Additional keyword arguments for the metric function.
    contamination : float in (0., 0.5), optional (default='auto')
        The amount of contamination of the data set, i.e. the proportion
        of outliers in the data set. When fitting this is used to define the
        threshold on the decision function. If "auto", the decision function
        threshold is determined as in the original paper [BKNS2000]_.
    novelty : boolean, default False
        By default, LocalOutlierFactor is only meant to be used for outlier
        detection (novelty=False). Set novelty to True if you want to use
        LocalOutlierFactor for novelty detection. In this case be aware that
        that you should only use predict, decision_function and score_samples
        on new unseen data and not on the training set.
    n_jobs : int or None, optional (default=None)
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
        Affects only :meth:`kneighbors` and :meth:`kneighbors_graph` methods.
    multivariate_metric : boolean, optional (default = False)
        Indicates if the metric used is a sklearn distance between vectors (see
        :class:`~sklearn.neighbors.DistanceMetric`) or a functional metric of
        the module `skfda.misc.metrics` if ``False``.

    Attributes
    ----------
    negative_outlier_factor_ : numpy array, shape (n_samples,)
        The opposite LOF of the training samples. The higher, the more normal.
        Inliers tend to have a LOF score close to 1
        (``negative_outlier_factor_`` close to -1), while outliers tend to have
        a larger LOF score.
        The local outlier factor (LOF) of a sample captures its
        supposed 'degree of abnormality'.
        It is the average of the ratio of the local reachability density of
        a sample and those of its k-nearest neighbors.
    n_neighbors_ : integer
        The actual number of neighbors used for :meth:`kneighbors` queries.
    offset_ : float
        Offset used to obtain binary labels from the raw scores.
        Observations having a negative_outlier_factor smaller than `offset_`
        are detected as abnormal.
        The offset is set to -1.5 (inliers score around -1), except when a
        contamination parameter different than "auto" is provided. In that
        case, the offset is defined in such a way we obtain the expected
        number of outliers in training.

    Examples:

        **Local Outlier Factor (LOF) for outlier detection**.

        >>> from skfda.exploratory.outliers import LocalOutlierFactor

        Creation of simulated dataset with 2 outliers to be used with LOF.

        >>> from skfda.datasets import make_sinusoidal_process
        >>> fd_clean = make_sinusoidal_process(n_samples=25, error_std=0,
        ...                                    phase_std=0.1, random_state=0)
        >>> fd_outliers = make_sinusoidal_process(
        ...     n_samples=2, error_std=0, phase_mean=0.5, random_state=5)
        >>> fd = fd_outliers.concatenate(fd_clean) # Dataset with 2 outliers

        Detection of outliers with LOF.

        >>> lof = LocalOutlierFactor()
        >>> is_outlier = lof.fit_predict(fd)
        >>> is_outlier # -1 for anomalies/outliers and +1 for inliers
        array([-1, -1,  1,  1,  1,  1,  1,  1, ...,  1,  1,  1,  1])

        The negative outlier factor stored.

        >>> lof.negative_outlier_factor_.round(2)
        array([-7.11, -1.54, -1.  , -0.99, ..., -0.97,  -1. ,  -0.99])

        **Novelty detection with LOF**.

        Creation of a dataset without outliers.

        >>> fd_train = make_sinusoidal_process(n_samples=25, error_std=0,
        ...                                    phase_std=0.1, random_state=9)

        Fit of LOF using the dataset without outliers.

        >>> lof = LocalOutlierFactor(novelty=True)
        >>> lof.fit(fd_train)
        LocalOutlierFactor(...novelty=True)

        Detection of annomalies for new samples.

        >>> lof.predict(fd) # Predict with samples not used in fit
        array([-1, -1,  1,  1,  1,  1,  1,  1, ...,  1,  1,  1,  1])


    References
    ----------
    .. [BKNS2000] Breunig, M. M., Kriegel, H. P., Ng, R. T., & Sander,
       J. (2000, May). LOF: identifying density-based local outliers. In ACM
       sigmod record.

    Notes
    -----
        This estimator wraps the scikit-learn class
        :class:`~sklearn.neighbors.LocalOutlierFactor` employing functional
        metrics and data instead of the multivariate ones.

    See also
    --------
    :class:`~skfda.ml.classification.KNeighborsClassifier`
    :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
    :class:`~skfda.ml.classification.NearestCentroids`
    :class:`~skfda.ml.regression.KNeighborsRegressor`
    :class:`~skfda.ml.regression.RadiusNeighborsRegressor`
    :class:`~skfda.ml.clustering.NearestNeighbors`
    """

    def __init__(self, n_neighbors=20, algorithm='auto',
                 leaf_size=30, metric='l2', metric_params=None,
                 contamination='auto', novelty=False,
                 n_jobs=1, multivariate_metric=False):
        """Initialize the Local Outlier Factor estimator."""

        super().__init__(n_neighbors=n_neighbors, algorithm=algorithm,
                         leaf_size=leaf_size, metric=metric,
                         metric_params=metric_params, n_jobs=n_jobs,
                         multivariate_metric=multivariate_metric)
        self.contamination = contamination
        self.novelty = novelty

    def _init_estimator(self, sklearn_metric):
        """Initialize the sklearn nearest neighbors estimator.

        Args:
            sklearn_metric: (pyfunc or 'precomputed'): Metric compatible with
                sklearn API or matrix (n_samples, n_samples) with precomputed
                distances.

        Returns:
            Sklearn LocalOutlierFactor estimator initialized.

        """
        from sklearn.neighbors import LocalOutlierFactor as _LocalOutlierFactor

        return _LocalOutlierFactor(
            n_neighbors=self.n_neighbors, algorithm=self.algorithm,
            leaf_size=self.leaf_size, metric=sklearn_metric,
            metric_params=self.metric_params, contamination=self.contamination,
            novelty=self.novelty, n_jobs=self.n_jobs)

    def _store_fit_data(self):
        """Store the parameters created during the fit."""
        self.negative_outlier_factor_ = self.estimator_.negative_outlier_factor_
        self.n_neighbors_ = self.estimator_.n_neighbors_
        self.offset_ = self.estimator_.offset_

    def fit(self, X, y=None):
        """Fit the model using X as training data.

        Parameters
        ----------
        X : :class:`~skfda.FDataGrid` or array_like
            Training data. FDataGrid containing the samples,
            or array with shape [n_samples, n_samples] if metric='precomputed'.
        y : Ignored
            not used, present for API consistency by convention.
        Returns
        -------
        self : object
        """

        super().fit(X, y)
        self._store_fit_data()

        return self

    def predict(self, X=None):
        """Predict the labels (1 inlier, -1 outlier) of X according to LOF.

        This method allows to generalize prediction to *new observations* (not
        in the training set). Only available for novelty detection (when
        novelty is set to True).

        If X is None, returns the same as fit_predict(X_train).

        Parameters
        ----------
        X : :class:`~skfda.FDataGrid` or array_like
            FDataGrid containing the query sample or samples to compute the
            Local Outlier Factor w.r.t. to the training samples or array with
            the distances to the training samples if metric='precomputed'.

        Returns
        -------
        is_inlier : array, shape (n_samples,)
            Returns -1 for anomalies/outliers and +1 for inliers.
        """

        self._check_is_fitted()
        X_multivariate = self._transform_to_multivariate(X)

        return self.estimator_.predict(X_multivariate)

    def fit_predict(self, X, y=None):
        """Fits the model to the training set X and returns the labels.

        Label is 1 for an inlier and -1 for an outlier according to the LOF
        score and the contamination parameter.

        Parameters
        ----------
        X : :class:`~skfda.FDataGrid` or array_like
            Training data. FDataGrid containing the samples,
            or array with shape [n_samples, n_samples] if metric='precomputed'.
        y : Ignored
            not used, present for API consistency by convention.
        Returns
        -------
        is_inlier : array, shape (n_samples,)
            Returns -1 for anomalies/outliers and 1 for inliers.
        """

        # In this estimator fit_predict cannot be wrapped as fit().predict()

        if self.metric == 'precomputed':
            self.estimator_ = self._init_estimator(self.metric)
            res = self.estimator_.fit_predict(X, y)
        else:
            self._grid_points = X.grid_points
            self._shape = X.data_matrix.shape[1:]

            if not self.multivariate_metric:
                # Constructs sklearn metric to manage vector
                if self.metric == 'l2':
                    metric = l2_distance
                else:
                    metric = self.metric
                sklearn_metric = _to_multivariate_metric(metric,
                                                         self._grid_points)
            else:
                sklearn_metric = self.metric

            self.estimator_ = self._init_estimator(sklearn_metric)
            X_multivariate = self._transform_to_multivariate(X)
            res = self.estimator_.fit_predict(X_multivariate, y)

        self._store_fit_data()

        return res

    def decision_function(self, X):
        """Shifted opposite of the Local Outlier Factor of X.

        Bigger is better, i.e. large values correspond to inliers.
        The shift offset allows a zero threshold for being an outlier.
        Only available for novelty detection (when novelty is set to True).
        The argument X is supposed to contain *new data*: if X contains a
        point from training, it considers the later in its own neighborhood.
        Also, the samples in X are not considered in the neighborhood of any
        point.

        Parameters
        ----------
        X : :class:`~skfda.FDataGrid` or array_like
            FDataGrid containing the query sample or samples to compute the
            Local Outlier Factor w.r.t. to the training samples.

        Returns
        -------
        shifted_opposite_lof_scores : array, shape (n_samples,)
            The shifted opposite of the Local Outlier Factor of each input
            samples. The lower, the more abnormal. Negative scores represent
            outliers, positive scores represent inliers.
        """
        self._check_is_fitted()
        X_multivariate = self._transform_to_multivariate(X)

        return self.estimator_.decision_function(X_multivariate)

    def score_samples(self, X):
        """Opposite of the Local Outlier Factor of X.

        It is the opposite as bigger is better, i.e. large values correspond
        to inliers.

        Only available for novelty detection (when novelty is set to True).
        The argument X is supposed to contain *new data*: if X contains a
        point from training, it considers the later in its own neighborhood.
        Also, the samples in X are not considered in the neighborhood of any
        point.

        The score_samples on training data is available by considering the
        the ``negative_outlier_factor_`` attribute.

        Parameters
        ----------
        X : :class:`~skfda.FDataGrid` or array_like
            FDataGrid containing the query sample or samples to compute the
            Local Outlier Factor w.r.t. to the training samples.

        Returns
        -------
        opposite_lof_scores : array, shape (n_samples,)
            The opposite of the Local Outlier Factor of each input samples.
            The lower, the more abnormal.
        """
        self._check_is_fitted()
        X_multivariate = self._transform_to_multivariate(X)

        return self.estimator_.score_samples(X_multivariate)



import itertools
from typing import Callable, Optional, Sequence, Union

import numpy as np
import scipy.integrate
import sklearn.utils
from scipy.stats import multivariate_normal

from .. import FDataGrid
from .._utils import (
    RandomStateLike,
    _cartesian_product,
    _to_grid_points,
    normalize_warping,
)
from ..misc import covariances
from ..representation._typing import (
    DomainRangeLike,
    GridPointsLike,
    NDArrayFloat,
)
from ..representation.interpolation import SplineInterpolation

MeanCallable = Callable[[np.ndarray], np.ndarray]
CovarianceCallable = Callable[[np.ndarray, np.ndarray], np.ndarray]

MeanLike = Union[float, np.ndarray, MeanCallable]
CovarianceLike = Union[None, np.ndarray, CovarianceCallable]


def make_gaussian(
    n_samples: int = 100,
    *,
    grid_points: GridPointsLike,
    domain_range: Optional[DomainRangeLike] = None,
    mean: MeanLike = 0,
    cov: CovarianceLike = None,
    noise: float = 0,
    random_state: RandomStateLike = None,
) -> FDataGrid:
    """
    Generate Gaussian random fields.

    Args:
        n_samples: The total number of trajectories.
        grid_points: Sample points for the evaluation grid of the
            Gaussian field.
        domain_range: The domain range of the returned functional
            observations.
        mean: The mean function of the random field. Can be a callable
            accepting a vector with the locations, or a vector with
            appropriate size.
        cov: The covariance function of the process. Can be a
            callable accepting two vectors with the locations, or a
            matrix with appropriate size. By default,
            the Brownian covariance function is used.
        noise: Standard deviation of Gaussian noise added to the data.
        random_state: Random state.

    Returns:
        :class:`FDataGrid` object comprising all the trajectories.

    See also:
        :func:`make_gaussian_process`: Simpler function for generating
        Gaussian processes.

    """
    random_state = sklearn.utils.check_random_state(random_state)

    if cov is None:
        cov = covariances.Brownian()

    grid_points = _to_grid_points(grid_points)

    input_points = _cartesian_product(grid_points)

    covariance = covariances._execute_covariance(
        cov, input_points, input_points)

    if noise:
        covariance += np.eye(len(covariance)) * noise ** 2

    mu = np.zeros(len(input_points))
    if callable(mean):
        mean = mean(input_points)

    mu += np.ravel(mean)

    data_matrix = random_state.multivariate_normal(
        mu.ravel(),
        covariance,
        n_samples,
    )

    data_matrix = data_matrix.reshape(
        [n_samples] + [len(t) for t in grid_points] + [-1],
    )

    return FDataGrid(
        grid_points=grid_points,
        data_matrix=data_matrix,
        domain_range=domain_range,
    )


def make_gaussian_process(
    n_samples: int = 100,
    n_features: int = 100,
    *,
    start: float = 0,
    stop: float = 1,
    mean: MeanLike = 0,
    cov: CovarianceLike = None,
    noise: float = 0,
    random_state: RandomStateLike = None,
) -> FDataGrid:
    """Generate Gaussian process trajectories.

    Args:
        n_samples: The total number of trajectories.
        n_features: The total number of features (points of evaluation).
        start: Starting point of the trajectories.
        stop: Ending point of the trajectories.
        mean: The mean function of the process. Can be a callable accepting
              a vector with the locations, or a vector with length
              ``n_features``.
        cov: The covariance function of the process. Can be a
              callable accepting two vectors with the locations, or a
              matrix with size ``n_features`` x ``n_features``. By default,
              the Brownian covariance function is used.
        noise: Standard deviation of Gaussian noise added to the data.
        random_state: Random state.

    Returns:
        :class:`FDataGrid` object comprising all the trajectories.

    See also:
        :func:`make_gaussian`: More general function that allows to
        select the points of evaluation and to
        generate data in higer dimensions.

    """
    t = np.linspace(start, stop, n_features)

    return make_gaussian(
        n_samples=n_samples,
        grid_points=[t],
        mean=mean,
        cov=cov,
        noise=noise,
        random_state=random_state,
    )


def make_sinusoidal_process(
    n_samples: int = 15,
    n_features: int = 100,
    *,
    start: float = 0,
    stop: float = 1,
    period: float = 1,
    phase_mean: float = 0,
    phase_std: float = 0.6,
    amplitude_mean: float = 1,
    amplitude_std: float = 0.05,
    error_std: float = 0.2,
    random_state: RandomStateLike = None,
) -> FDataGrid:
    r"""Generate sinusoidal proccess.

    Each sample :math:`x_i(t)` is generated as:

    .. math::

        x_i(t) = \alpha_i \sin(\omega t + \phi_i) + \epsilon_i(t)


    where :math:`\omega=\frac{2 \pi}{\text{period}}`. Amplitudes
    :math:`\alpha_i` and phases :math:`\phi_i` are normally distributed.
    :math:`\epsilon_i(t)` is a gaussian white noise process.

    Args:
        n_samples: Total number of samples.
        n_features: Points per sample.
        start: Starting point of the samples.
        stop: Ending point of the samples.
        period: Period of the sine function.
        phase_mean: Mean of the phase.
        phase_std: Standard deviation of the phase.
        amplitude_mean: Mean of the amplitude.
        amplitude_std: Standard deviation of the amplitude.
        error_std: Standard deviation of the gaussian Noise.
        random_state: Random state.

    Returns:
        :class:`FDataGrid` object comprising all the samples.

    """
    random_state = sklearn.utils.check_random_state(random_state)

    t = np.linspace(start, stop, n_features)

    alpha = np.diag(random_state.normal(
        amplitude_mean,
        amplitude_std,
        n_samples,
    ))

    phi = np.outer(
        random_state.normal(phase_mean, phase_std, n_samples),
        np.ones(n_features),
    )

    error = random_state.normal(0, error_std, (n_samples, n_features))

    y = alpha @ np.sin((2 * np.pi / period) * t + phi) + error

    return FDataGrid(grid_points=t, data_matrix=y)


def make_multimodal_landmarks(
    n_samples: int = 15,
    *,
    n_modes: int = 1,
    dim_domain: int = 1,
    dim_codomain: int = 1,
    start: float = -1,
    stop: float = 1,
    std: float = 0.05,
    random_state: RandomStateLike = None,
) -> np.ndarray:
    """Generate landmarks points.

    Used by :func:`make_multimodal_samples` to generate the location of the
    landmarks.

    Generates a matrix containing the landmarks or locations of the modes
    of the samples generates by :func:`make_multimodal_samples`.

    If the same random state is used when generating the landmarks and
    multimodal samples, these will correspond to the position of the modes of
    the multimodal samples.

    Args:
        n_samples: Total number of samples.
        n_modes: Number of modes of each sample.
        dim_domain: Number of dimensions of the domain.
        dim_codomain: Number of dimensions of the codomain.
        start: Starting point of the samples. In multidimensional objects the
            starting point of the axis.
        stop: Ending point of the samples. In multidimensional objects the
            ending point of the axis.
        std: Standard deviation of the variation of the modes location.
        random_state: Random state.

    Returns:
        :class:`np.ndarray` with the location of the modes, where the component
        (i,j,k) corresponds to the mode k of the image dimension j of the
        sample i.

    """
    random_state = sklearn.utils.check_random_state(random_state)

    modes_location = np.linspace(start, stop, n_modes + 2)[1:-1]
    modes_location = np.repeat(
        modes_location[:, np.newaxis],
        dim_domain,
        axis=1,
    )

    variation = random_state.multivariate_normal(
        (0,) * dim_domain,
        std * np.eye(dim_domain),
        size=(n_samples, dim_codomain, n_modes),
    )

    return modes_location + variation


def make_multimodal_samples(
    n_samples: int = 15,
    *,
    n_modes: int = 1,
    points_per_dim: int = 100,
    dim_domain: int = 1,
    dim_codomain: int = 1,
    start: float = -1,
    stop: float = 1,
    std: float = 0.05,
    mode_std: float = 0.02,
    noise: float = 0,
    modes_location: Optional[Union[Sequence[float], NDArrayFloat]] = None,
    random_state: RandomStateLike = None,
) -> FDataGrid:
    r"""
    Generate multimodal samples.

    Each sample :math:`x_i(t)` is proportional to a gaussian mixture, generated
    as the sum of multiple pdf of multivariate normal distributions with
    different means.

    .. math::

        x_i(t) \propto \sum_{n=1}^{\text{n\_modes}} \exp \left (
        {-\frac{1}{2\sigma} (t-\mu_n)^T \mathbb{1} (t-\mu_n)} \right )

    Where :math:`\mu_n=\text{mode\_location}_n+\epsilon` and :math:`\epsilon`
    is normally distributed, with mean :math:`\mathbb{0}` and standard
    deviation given by the parameter `std`.

    Args:
        n_samples: Total number of samples.
        n_modes: Number of modes of each sample.
        points_per_dim: Points per sample. If the object is multidimensional
            indicates the number of points for each dimension in the domain.
            The sample will have :math:
            `\text{points_per_dim}^\text{dim_domain}` points of
            discretization.
        dim_domain: Number of dimensions of the domain.
        dim_codomain: Number of dimensions of the image
        start: Starting point of the samples. In multidimensional objects the
            starting point of each axis.
        stop: Ending point of the samples. In multidimensional objects the
            ending point of each axis.
        std: Standard deviation of the variation of the modes location.
        mode_std: Standard deviation :math:`\sigma` of each mode.
        noise: Standard deviation of Gaussian noise added to the data.
        modes_location:  List of coordinates of each mode.
        random_state: Random state.

    Returns:
        :class:`FDataGrid` object comprising all the samples.

    """
    random_state = sklearn.utils.check_random_state(random_state)

    if modes_location is None:

        location = make_multimodal_landmarks(
            n_samples=n_samples,
            n_modes=n_modes,
            dim_domain=dim_domain,
            dim_codomain=dim_codomain,
            start=start,
            stop=stop,
            std=std,
            random_state=random_state,
        )

    else:

        location = np.asarray(modes_location).reshape(
            (n_samples, dim_codomain, n_modes, dim_domain),
        )

    axis = np.linspace(start, stop, points_per_dim)

    if dim_domain == 1:
        grid_points = axis
        evaluation_grid = axis
    else:
        grid_points = np.repeat(axis[:, np.newaxis], dim_domain, axis=1).T

        meshgrid = np.meshgrid(*grid_points)

        evaluation_grid = np.empty(meshgrid[0].shape + (dim_domain,))

        for i in range(dim_domain):
            evaluation_grid[..., i] = meshgrid[i]

    # Data matrix of the grid
    shape = (n_samples,) + dim_domain * (points_per_dim,) + (dim_codomain,)
    data_matrix = np.zeros(shape)

    # Covariance matrix of the samples
    cov = mode_std * np.eye(dim_domain)

    for i, j, k in itertools.product(
        range(n_samples),
        range(dim_codomain),
        range(n_modes),
    ):
        data_matrix[i, ..., j] += multivariate_normal.pdf(
            evaluation_grid,
            location[i, j, k],
            cov,
        )

    # Constant to make modes value aprox. 1
    data_matrix *= (2 * np.pi * mode_std) ** (dim_domain / 2)

    data_matrix += random_state.normal(0, noise, size=data_matrix.shape)

    return FDataGrid(grid_points=grid_points, data_matrix=data_matrix)


def make_random_warping(
    n_samples: int = 15,
    n_features: int = 100,
    *,
    start: float = 0,
    stop: float = 1,
    sigma: float = 1,
    shape_parameter: float = 50,
    n_random: int = 4,
    random_state: RandomStateLike = None,
) -> FDataGrid:
    r"""Generate random warping functions.

    Let :math:`v(t)` be a randomly generated function defined in :math:`[0,1]`

    .. math::
        sv(t) = \sum_{j=0}^{N} a_j \sin(\frac{2 \pi j}{K}t) + b_j
        \cos(\frac{2 \pi j}{K}t)

    where :math:`a_j, b_j \sim N(0, \sigma)`.

    The random warping it is constructed making an exponential map to
    :math:`\Gamma`.

     .. math::
        \gamma(t) = \int_0^t \left ( \frac{\sin(\|v\|)}{\|v\|} v(s) +
        \cos(\|v\|) \right )^2 ds

    An affine traslation it is used to define the warping in :math:`[a,b]`.

    The smoothing and shape of the warpings can be controlling changing
    :math:`N`, :math:`\sigma` and :math:`K= 1 + \text{shape\_parameter}`.


    Args:
        n_samples: Total number of samples. Defaults 15.
        n_features: The total number of trajectories. Defaults 100.
        start: Starting point of the samples. Defaults 1.
        stop: Ending point of the samples. Defaults 0.
        sigma: Parameter to control the variance of the samples. Defaults 1.
        shape_parameter: Parameter to control the shape of the warpings.
            Should be a positive value. When the shape parameter goes to
            infinity the warpings generated are :math:`\gamma_{id}`.
            Defaults to 50.
        n_random: Number of random sines and cosines to be sum to construct
            the warpings.
        random_state: Random state.

    Returns:
        Object comprising all the samples.

    """
    # Based on the original implementation of J. D. Tucker in the
    # package python_fdasrsf <https://github.com/jdtuck/fdasrsf_python>.

    random_state = sklearn.utils.check_random_state(random_state)

    freq = shape_parameter + 1

    # Frequency
    omega = 2 * np.pi / freq
    sqrt2 = np.sqrt(2)
    sqrt_sigma = np.sqrt(sigma)

    # Originally it is compute in (0,1), then it is rescaled
    time = np.outer(np.linspace(0, 1, n_features), np.ones(n_samples))

    # Operates trasposed to broadcast dimensions
    v = np.outer(
        np.ones(n_features),
        random_state.normal(scale=sqrt_sigma, size=n_samples),
    )

    for j in range(2, 2 + n_random):
        alpha = random_state.normal(scale=sqrt_sigma, size=(2, n_samples))
        alpha *= sqrt2
        v += alpha[0] * np.cos(j * omega * time)
        v += alpha[1] * np.sin(j * omega * time)

    v -= v.mean(axis=0)

    # Equivalent to RN norm (equispaced monitorization points)
    v_norm = np.linalg.norm(v, axis=0)
    v_norm /= np.sqrt(n_features)

    # Exponential map of vectors
    v *= np.sin(v_norm) / v_norm
    v += np.cos(v_norm)
    np.square(v, out=v)

    # Creation of FDataGrid in the corresponding domain
    data_matrix = scipy.integrate.cumtrapz(
        v,
        dx=1 / n_features,
        initial=0,
        axis=0,
    )
    warping = FDataGrid(data_matrix.T, grid_points=time[:, 0])
    warping = normalize_warping(warping, domain_range=(start, stop))
    warping.interpolation = SplineInterpolation(
        interpolation_order=3,
        monotone=True,
    )

    return warping



"""BasePlot Module.

This module contains the abstract class of which inherit all
the visualization modules, containing the basic functionality
common to all of them.
"""

from abc import ABC, abstractmethod
from typing import Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.backend_bases import LocationEvent, MouseEvent
from matplotlib.collections import PathCollection
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure
from matplotlib.text import Annotation

from ...representation import FData
from ...representation._typing import NDArrayInt
from ._utils import _figure_to_svg, _get_figure_and_axes, _set_figure_layout


class BasePlot(ABC):
    """
    BasePlot class.

    Attributes:
        artists: List of Artist objects corresponding
            to every instance of our plot. They will be used to modify
            the visualization with interactivity and widgets.
        fig: Figure over with the graphs are plotted.
        axes: Sequence of axes where the graphs are plotted.
    """

    @abstractmethod
    def __init__(
        self,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Union[Axes, Sequence[Axes], None] = None,
        n_rows: Optional[int] = None,
        n_cols: Optional[int] = None,
        c: Optional[NDArrayInt] = None,
        cmap_bold: ListedColormap = None,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
    ) -> None:
        self.artists: Optional[np.ndarray] = None
        self.chart = chart
        self.fig = fig
        self.axes = axes
        self.n_rows = n_rows
        self.n_cols = n_cols
        self._tag = self._create_annotation()
        self.c = c
        self.cmap_bold = cmap_bold
        self.x_label = x_label
        self.y_label = y_label

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:
        pass

    def plot(
        self,
    ) -> Figure:
        """
        Plot the object and its data.

        Returns:
            Figure: figure object in which the displays and
                widgets will be plotted.
        """
        fig = getattr(self, "fig_", None)
        axes = getattr(self, "axes_", None)

        if fig is None:
            fig, axes = self._set_figure_and_axes(
                self.chart,
                fig=self.fig,
                axes=self.axes,
            )
        if self.x_label is not None:
            axes[0].set_xlabel(self.x_label)
        if self.y_label is not None:
            axes[0].set_ylabel(self.y_label)

        self._plot(fig, axes)

        self._hover_event_id = fig.canvas.mpl_connect(
            'motion_notify_event',
            self.hover,
        )

        return fig

    @property
    def dim(self) -> int:
        """Get the number of dimensions for this plot."""
        return 2

    @property
    def n_subplots(self) -> int:
        """Get the number of subplots that this plot uses."""
        return 1

    @property
    def n_samples(self) -> Optional[int]:
        """Get the number of instances that will be used for interactivity."""
        return None

    def _set_figure_and_axes(
        self,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Union[Axes, Sequence[Axes], None] = None,
    ) -> Tuple[Figure, Sequence[Axes]]:
        fig, axes = _get_figure_and_axes(chart, fig, axes)
        fig, axes = _set_figure_layout(
            fig=fig,
            axes=axes,
            dim=self.dim,
            n_axes=self.n_subplots,
            n_rows=self.n_rows,
            n_cols=self.n_cols,
        )

        self.fig_ = fig
        self.axes_ = axes

        return fig, axes

    def _repr_svg_(self) -> str:
        """Automatically represents the object as an svg when calling it."""
        self.fig = self.plot()
        plt.close(self.fig)
        return _figure_to_svg(self.fig)

    def _create_annotation(self) -> Annotation:
        tag = Annotation(
            "",
            xy=(0, 0),
            xytext=(20, 20),
            textcoords="offset points",
            bbox={
                "boxstyle": "round",
                "fc": "w",
            },
            arrowprops={
                "arrowstyle": "->",
            },
            annotation_clip=False,
            clip_on=False,
        )

        tag.get_bbox_patch().set_facecolor(color='khaki')
        intensity = 0.8
        tag.get_bbox_patch().set_alpha(intensity)

        return tag

    def _update_annotation(
        self,
        tag: Annotation,
        *,
        axes: Axes,
        sample_number: int,
        fdata: Optional[FData],
        position: Tuple[float, float],
    ) -> None:
        """
        Auxiliary method used to update the hovering annotations.

        Method used to update the annotations that appear while
        hovering a scattered point. The annotations indicate
        the index and coordinates of the point hovered.
        Args:
            tag: Annotation to update.
            axes: Axes were the annotation belongs.
            sample_number: Number of the current sample.
        """
        xdata_graph, ydata_graph = position

        tag.xy = (xdata_graph, ydata_graph)

        sample_name = (
            fdata.sample_names[sample_number]
            if fdata is not None
            else None
        )

        sample_descr = f" ({sample_name})" if sample_name is not None else ""

        text = (
            f"{sample_number}{sample_descr}: "
            f"({xdata_graph:.3g}, {ydata_graph:.3g})"
        )
        tag.set_text(text)

        x_axis = axes.get_xlim()
        y_axis = axes.get_ylim()

        label_xpos = -60
        label_ypos = 20
        if (xdata_graph - x_axis[0]) > (x_axis[1] - xdata_graph):
            label_xpos = -80

        if (ydata_graph - y_axis[0]) > (y_axis[1] - ydata_graph):
            label_ypos = -20

        if tag.figure:
            tag.remove()
        tag.figure = None
        axes.add_artist(tag)
        tag.set_transform(axes.transData)
        tag.set_position((label_xpos, label_ypos))

    def _sample_artist_from_event(
        self,
        event: LocationEvent,
    ) -> Optional[Tuple[int, Optional[FData], Artist]]:
        """Get the number, fdata and artist under a location event."""
        if self.artists is None:
            return None

        try:
            i = self.axes_.index(event.inaxes)
        except ValueError:
            return None

        for j, artist in enumerate(self.artists[:, i]):
            if not isinstance(artist, PathCollection):
                return None

            if artist.contains(event)[0]:
                return j, getattr(self, "fdata", None), artist

        return None

    def hover(self, event: MouseEvent) -> None:
        """
        Activate the annotation when hovering a point.

        Callback method that activates the annotation when hovering
        a specific point in a graph. The annotation is a description
        of the point containing its coordinates.
        Args:
            event: event object containing the artist of the point
                hovered.

        """
        found_artist = self._sample_artist_from_event(event)

        if event.inaxes is not None and found_artist is not None:
            sample_number, fdata, artist = found_artist

            self._update_annotation(
                self._tag,
                axes=event.inaxes,
                sample_number=sample_number,
                fdata=fdata,
                position=artist.get_offsets()[0],
            )
            self._tag.set_visible(True)
            self.fig_.canvas.draw_idle()
        elif self._tag.get_visible():
            self._tag.set_visible(False)
            self.fig_.canvas.draw_idle()



"""Fisher-Rao elastic registration."""
from __future__ import annotations

import warnings
from typing import Callable, Optional, TypeVar, Union

from sklearn.utils.validation import check_is_fitted

from ... import FDataGrid
from ..._utils import (
    _check_compatible_fdatagrid,
    check_is_univariate,
    invert_warping,
    normalize_scale,
)
from ...exploratory.stats import fisher_rao_karcher_mean
from ...exploratory.stats._fisher_rao import _elastic_alignment_array
from ...misc.operators import SRSF
from ...representation._typing import ArrayLike
from ...representation.interpolation import SplineInterpolation
from .base import InductiveRegistrationTransformer

_MeanType = Callable[[FDataGrid], FDataGrid]
SelfType = TypeVar("SelfType", bound="FisherRaoElasticRegistration")


class FisherRaoElasticRegistration(
    InductiveRegistrationTransformer[FDataGrid, FDataGrid],
):
    r"""Align a FDatagrid using the SRSF framework.

    Let :math:`f` be a function of the functional data object wich will be
    aligned to the template :math:`g`. Calculates the warping wich minimises
    the Fisher-Rao distance between :math:`g` and the registered function
    :math:`f^*(t)=f(\gamma^*(t))=f \circ \gamma^*`.

    .. math::
        \gamma^* = argmin_{\gamma \in \Gamma} d_{\lambda}(f \circ
        \gamma, g)

    Where :math:`d_{\lambda}` denotes the extended Fisher-Rao distance with a
    penalty term, used to control the amount of warping.

    .. math::
        d_{\lambda}^2(f \circ \gamma, g) = \| SRSF(f \circ \gamma)
        \sqrt{\dot{\gamma}} - SRSF(g)\|_{\mathbb{L}^2}^2 + \lambda
        \mathcal{R}(\gamma)

    In the implementation it is used as penalty term

    .. math::
        \mathcal{R}(\gamma) = \|\sqrt{\dot{\gamma}}- 1 \|_{\mathbb{L}^2}^2

    Wich restrict the amount of elasticity employed in the alignment.

    The registered function :math:`f^*(t)` can be calculated using the
    composition :math:`f^*(t)=f(\gamma^*(t))`.

    If the template is not specified it is used the Karcher mean of the set of
    functions under the elastic metric to perform the alignment, also known as
    `elastic mean`, wich is the local minimum of the sum of squares of elastic
    distances. See :func:`~elastic_mean`.

    In :footcite:`srivastava+klassen_2016_analysis_elastic` are described
    extensively the algorithms employed and the SRSF framework.

    Args:
        template: Template to
            align the curves. Can contain 1 sample to align all the curves to
            it or the same number of samples than the fdatagrid. By default
            `elastic mean`, in which case :func:`elastic_mean` is called.
        penalty_term: Controls the amount of elasticity.
            Defaults to 0.
        output_points: Set of points where the
            functions are evaluated, by default uses the sample points of the
            fdatagrid which will be transformed.
        grid_dim: Dimension of the grid used in the DP
            alignment algorithm. Defaults 7.

    Attributes:
        template\_: Template learned during fitting,
            used for alignment in :meth:`transform`.
        warping\_: Warping applied during the last
            transformation.

    References:
        .. footbibliography::

    Examples:
        Elastic registration of with train/test sets.

        >>> from skfda.preprocessing.registration import (
        ...     FisherRaoElasticRegistration,
        ... )
        >>> from skfda.datasets import make_multimodal_samples
        >>> X_train = make_multimodal_samples(n_samples=15, random_state=0)
        >>> X_test = make_multimodal_samples(n_samples=3, random_state=1)

        Fit the transformer, which learns the elastic mean of the train
        set as template.

        >>> elastic_registration = FisherRaoElasticRegistration()
        >>> elastic_registration.fit(X_train)
        FisherRaoElasticRegistration(...)

        Registration of the test set.

        >>> elastic_registration.transform(X_test)
        FDataGrid(...)

    """

    def __init__(
        self,
        *,
        template: Union[FDataGrid, _MeanType] = fisher_rao_karcher_mean,
        penalty: float = 0,
        output_points: Optional[ArrayLike] = None,
        grid_dim: int = 7,
    ) -> None:
        self.template = template
        self.penalty = penalty
        self.output_points = output_points
        self.grid_dim = grid_dim

    def fit(self: SelfType, X: FDataGrid, y: None = None) -> SelfType:

        # Points of discretization
        if self.output_points is None:
            self._output_points = X.grid_points[0]
        else:
            self._output_points = self.output_points

        if isinstance(self.template, FDataGrid):
            self.template_ = self.template  # Template already constructed
        else:
            self.template_ = self.template(X)

        _check_compatible_fdatagrid(X, self.template_)

        # Constructs the SRSF of the template
        srsf = SRSF(output_points=self._output_points, initial_value=0)
        self._template_srsf = srsf.fit_transform(self.template_)

        return self

    def transform(self, X: FDataGrid, y: None = None) -> FDataGrid:

        check_is_fitted(self)
        check_is_univariate(X)
        _check_compatible_fdatagrid(X, self.template_)

        if (
            len(self._template_srsf) != 1
            and len(X) != len(self._template_srsf)
        ):

            raise ValueError(
                "The template should contain one sample to align "
                "all the curves to the same function or the "
                "same number of samples than X.",
            )

        srsf = SRSF(output_points=self.output_points, initial_value=0)
        fdatagrid_srsf = srsf.fit_transform(X)

        # Points of discretization
        if self.output_points is None:
            output_points = fdatagrid_srsf.grid_points[0]
        else:
            output_points = self.output_points

        # Discretizacion in evaluation points
        q_data = fdatagrid_srsf(output_points)[..., 0]
        template_data = self._template_srsf(output_points)[..., 0]

        if q_data.shape[0] == 1:
            q_data = q_data[0]

        if template_data.shape[0] == 1:
            template_data = template_data[0]

        # Values of the warping
        gamma = _elastic_alignment_array(
            template_data,
            q_data,
            normalize_scale(output_points),
            self.penalty,
            self.grid_dim,
        )

        # Normalize warping to original interval
        gamma = normalize_scale(
            gamma,
            a=output_points[0],
            b=output_points[-1],
        )

        # Interpolation
        interpolation = SplineInterpolation(
            interpolation_order=3,
            monotone=True,
        )

        self.warping_ = FDataGrid(
            gamma,
            output_points,
            interpolation=interpolation,
        )

        return X.compose(self.warping_, eval_points=output_points)

    def inverse_transform(self, X: FDataGrid, y: None = None) -> FDataGrid:
        r"""
        Reverse the registration procedure previosly applied.

        Let :math:`gamma(t)` the warping applied to construct a registered
        functional datum :math:`f^*(t)=f(\gamma(t))`.

        Given a functional datum :math:`f^*(t) it is computed
        :math:`\gamma^{-1}(t)` to reverse the registration procedure
        :math:`f(t)=f^*(\gamma^{-1}(t))`.

        Args:
            X: Functional data to apply the reverse
                transform.
            y: Present for API conventions.

        Returns:
            Functional data compose by the inverse warping.

        Raises:
            ValueError: If the warpings :math:`\gamma` were not build via
                :meth:`transform` or if the number of samples of `X` is
                different than the number of samples of the dataset
                previously transformed.

        Examples:
            Center the datasets taking into account the misalignment.

            >>> from skfda.preprocessing.registration import (
            ...     FisherRaoElasticRegistration,
            ... )
            >>> from skfda.datasets import make_multimodal_samples
            >>> X = make_multimodal_samples(random_state=0)

            Registration of the dataset.

            >>> elastic_registration = FisherRaoElasticRegistration()
            >>> X = elastic_registration.fit_transform(X)

            Substract the elastic mean build as template during the
            registration and reverse the transformation.

            >>> X = X - elastic_registration.template_
            >>> X_center = elastic_registration.inverse_transform(X)
            >>> X_center
            FDataGrid(...)


        See also:
            :func:`invert_warping`

        """
        warping = getattr(self, 'warping_', None)

        if warping is None:
            raise ValueError(
                "Data must be previosly transformed to apply the "
                "inverse transform",
            )
        elif len(X) != len(warping):
            raise ValueError(
                "Data must contain the same number of samples "
                "than the dataset previously transformed",
            )

        inverse_warping = invert_warping(warping)

        return X.compose(inverse_warping, eval_points=self.output_points)


class ElasticRegistration(FisherRaoElasticRegistration):
    """Deprecated name for FisherRaoElasticRegistration."""

    def __init__(
        self,
        template: Union[FDataGrid, _MeanType] = fisher_rao_karcher_mean,
        penalty: float = 0,
        output_points: Optional[ArrayLike] = None,
        grid_dim: int = 7,
    ) -> None:
        warnings.warn(
            "ElasticRegistration has been renamed. "
            "Use FisherRaoElasticRegistration instead.",
            DeprecationWarning,
        )
        super().__init__(
            template=template,
            penalty=penalty,
            output_points=output_points,
            grid_dim=grid_dim,
        )



"""Feature extraction transformers for dimensionality reduction."""
from __future__ import annotations

from typing import Generic, Sequence, TypeVar, Union

import numpy as np
from numpy import ndarray
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted as sklearn_check_is_fitted

from ...._utils import _classifier_fit_depth_methods
from ....exploratory.depth import Depth, ModifiedBandDepth
from ....representation.grid import FData

T = TypeVar("T", bound=FData)


class DDGTransformer(
    BaseEstimator,  # type: ignore
    TransformerMixin,  # type: ignore
    Generic[T],
):
    r"""Generalized depth-versus-depth (DD) transformer for functional data.

    This transformer takes a list of k depths and performs the following map:

    .. math::
        \mathcal{X} &\rightarrow \mathbb{R}^G \\
        x &\rightarrow \textbf{d} = (D_1^1(x), D_1^2(x),...,D_g^k(x))

    Where :math:`D_i^j(x)` is the depth of the point :math:`x` with respect to
    the data in the :math:`i`-th group using the :math:`j`-th depth of the
    provided list.

    Note that :math:`\mathcal{X}` is possibly multivariate, that is,
    :math:`\mathcal{X} = \mathcal{X}_1 \times ... \times \mathcal{X}_p`.

    Parameters:
        depth_method:
            The depth class or sequence of depths to use when calculating
            the depth of a test sample in a class. See the documentation of
            the depths module for a list of available depths. By default it
            is ModifiedBandDepth.

    Examples:
        Firstly, we will import and split the Berkeley Growth Study dataset

        >>> from skfda.datasets import fetch_growth
        >>> from sklearn.model_selection import train_test_split
        >>> dataset = fetch_growth()
        >>> fd = dataset['data']
        >>> y = dataset['target']
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     fd, y, test_size=0.25, stratify=y, random_state=0)

        >>> from skfda.preprocessing.dim_reduction.feature_extraction import \
        ... DDGTransformer
        >>> from sklearn.pipeline import make_pipeline
        >>> from sklearn.neighbors import KNeighborsClassifier

        We classify by first transforming our data using the defined map
        and then using KNN

        >>> pipe = make_pipeline(DDGTransformer(), KNeighborsClassifier())
        >>> pipe.fit(X_train, y_train)
        Pipeline(steps=[('ddgtransformer',
                         DDGTransformer(depth_method=[ModifiedBandDepth()])),
                        ('kneighborsclassifier', KNeighborsClassifier())])

        We can predict the class of new samples

        >>> pipe.predict(X_test)
        array([1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1,
               1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])

        Finally, we calculate the mean accuracy for the test data

        >>> pipe.score(X_test, y_test)
        0.875

    References:
        Cuesta-Albertos, J. A., Febrero-Bande, M. and Oviedo de la Fuente, M.
        (2017). The DDG-classifier in the functional setting.
        TEST, 26. 119-142.
    """

    def __init__(  # noqa: WPS234
        self,
        depth_method: Union[Depth[T], Sequence[Depth[T]], None] = None,
    ) -> None:
        self.depth_method = depth_method

    def fit(self, X: T, y: ndarray) -> DDGTransformer[T]:
        """Fit the model using X as training data and y as target values.

        Args:
            X: FDataGrid with the training data.
            y: Target values of shape = (n_samples).

        Returns:
            self
        """
        if self.depth_method is None:
            self.depth_method = ModifiedBandDepth()

        if isinstance(self.depth_method, Depth):
            self.depth_method = [self.depth_method]

        classes, class_depth_methods = _classifier_fit_depth_methods(
            X, y, self.depth_method,
        )

        self._classes = classes
        self.class_depth_methods_ = class_depth_methods

        return self

    def transform(self, X: T) -> ndarray:
        """Transform the provided data using the defined map.

        Args:
            X: FDataGrid with the test samples.

        Returns:
            Array of shape (n_samples, G).
        """
        sklearn_check_is_fitted(self)

        return np.transpose([
            depth_method.transform(X)
            for depth_method in self.class_depth_methods_
        ])



"""Functional Data Boxplot Module.

This module contains the classes to construct the functional data boxplot and
visualize it.

"""
from __future__ import annotations

import math
from abc import abstractmethod
from typing import Optional, Sequence, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from matplotlib.figure import Figure

from skfda.exploratory.depth.multivariate import Depth

from ... import FData, FDataGrid
from ...representation._typing import NDArrayBool, NDArrayFloat
from ..depth import ModifiedBandDepth
from ..outliers import _envelopes
from ._baseplot import BasePlot
from ._utils import _set_labels


class FDataBoxplot(BasePlot):
    """
    Abstract class inherited by the Boxplot and SurfaceBoxplot classes.

    It the data of the functional boxplot or surface boxplot of a FDataGrid
    object, depending on the dimensions of the :term:`domain`, 1 or 2
    respectively.

    It forces to both classes, Boxplot and SurfaceBoxplot to conain at least
    the median, central and outlying envelopes and a colormap for their
    graphical representation, obtained calling the plot method.

    """

    @abstractmethod
    def __init__(
        self,
        chart: Union[Figure, Axes, None] = None,
        *,
        factor: float = 1.5,
        fig: Optional[Figure] = None,
        axes: Optional[Axes] = None,
        n_rows: Optional[int] = None,
        n_cols: Optional[int] = None,
    ) -> None:
        if factor < 0:
            raise ValueError(
                "The number used to calculate the "
                "outlying envelope must be positive.",
            )

        super().__init__(
            chart,
            fig=fig,
            axes=axes,
            n_rows=n_rows,
            n_cols=n_cols,
        )
        self._factor = factor

    @property
    def factor(self) -> float:
        return self._factor

    @property
    def fdatagrid(self) -> FDataGrid:
        pass

    @property
    def median(self) -> NDArrayFloat:
        pass

    @property
    def central_envelope(self) -> Tuple[NDArrayFloat, NDArrayFloat]:
        pass

    @property
    def non_outlying_envelope(self) -> Tuple[NDArrayFloat, NDArrayFloat]:
        pass

    @property
    def colormap(self) -> Colormap:
        return self._colormap

    @colormap.setter
    def colormap(self, value: Colormap) -> None:
        if not isinstance(value, matplotlib.colors.LinearSegmentedColormap):
            raise ValueError(
                "colormap must be of type "
                "matplotlib.colors.LinearSegmentedColormap",
            )
        self._colormap = value


class Boxplot(FDataBoxplot):
    r"""
    Representation of the functional boxplot.

    Class implementing the functionl boxplot which is an informative
    exploratory tool for visualizing functional data, as well as its
    generalization, the enhanced functional boxplot. Only supports 1
    dimensional :term:`domain` functional data.

    Based on the center outward ordering induced by a :ref:`depth measure
    <depth-measures>` for functional data, the descriptive statistics of a
    functional boxplot are: the envelope of the 50% central region, the median
    curve,and the maximum non-outlying envelope. In addition, outliers can be
    detected in a functional boxplot by the 1.5 times the 50% central region
    empirical rule, analogous to the rule for classical boxplots.

    For more information see :footcite:ts:`sun+genton_2011_boxplots`.

    Args:
        fdatagrid: Object containing the data.
        chart: figure over with the graphs are plotted or axis over
            where the graphs are plotted. If None and ax is also
            None, the figure is initialized.
        depth_method: Method used to order the data. Defaults to
            :func:`~skfda.exploratory.depth.ModifiedBandDepth`.
        prob: List with float numbers (in the range from 1 to 0) that
            indicate which central regions to represent.
            Defaults to (0.5,) which represents the 50% central region.
        factor: Number used to calculate the outlying envelope.
        fig: Figure over with the graphs are
            plotted in case ax is not specified. If None and ax is also
            None, the figure is initialized.
        axes: Axis over where the graphs
            are plotted. If None, see param fig.
        n_rows: Designates the number of rows of the figure
            to plot the different dimensions of the image. Only specified
            if fig and ax are None.
        n_cols: Designates the number of columns of the
            figure to plot the different dimensions of the image. Only
            specified if fig and ax are None.

    Attributes:
        fdatagrid: Object containing the data.
        median: Contains the median/s.
        central_envelope: Contains the central envelope/s.
        non_outlying_envelope: Contains the non-outlying envelope/s.
        colormap: Colormap from which the colors to represent the
            central regions are selected.
        envelopes: Contains the region envelopes.
        outliers: Contains the outliers.
        barcol: Color of the envelopes and vertical lines.
        outliercol: Color of the ouliers.
        mediancol: Color of the median.
        show_full_outliers: If False (the default) then only the part
            outside the box is plotted. If True, complete outling curves are
            plotted.

    Representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.datasets import make_gaussian_process
        from skfda.misc.covariances import Exponential
        from skfda.exploratory.visualization import Boxplot

        fd = make_gaussian_process(
                n_samples=20, cov=Exponential(), random_state=3)

        Boxplot(fd)


    Examples:
        Function :math:`f : \mathbb{R}\longmapsto\mathbb{R}`.

        >>> from skfda import FDataGrid
        >>> from skfda.exploratory.visualization import Boxplot
        >>>
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, grid_points, dataset_name="dataset",
        ...                argument_names=["x_label"],
        ...                coordinate_names=["y_label"])
        >>> Boxplot(fd)
        Boxplot(
            FDataGrid=FDataGrid(
                array([[[ 1. ],
                        [ 1. ],
                        [ 2. ],
                        [ 3. ],
                        [ 2.5],
                        [ 2. ]],
                       [[ 0.5],
                        [ 0.5],
                        [ 1. ],
                        [ 2. ],
                        [ 1.5],
                        [ 1. ]],
                       [[-1. ],
                        [-1. ],
                        [-0.5],
                        [ 1. ],
                        [ 1. ],
                        [ 0.5]],
                       [[-0.5],
                        [-0.5],
                        [-0.5],
                        [-1. ],
                        [-1. ],
                        [-1. ]]]),
                grid_points=(array([ 0.,  2.,  4.,  6.,  8., 10.]),),
                domain_range=((0.0, 10.0),),
                dataset_name='dataset',
                argument_names=('x_label',),
                coordinate_names=('y_label',),
                ...),
            median=array([[ 0.5],
                          [ 0.5],
                          [ 1. ],
                          [ 2. ],
                          [ 1.5],
                          [ 1. ]]),
            central envelope=(array([[-1. ],
                                     [-1. ],
                                     [-0.5],
                                     [ 1. ],
                                     [ 1. ],
                                     [ 0.5]]), array([[ 0.5],
                                     [ 0.5],
                                     [ 1. ],
                                     [ 2. ],
                                     [ 1.5],
                                     [ 1. ]])),
            non-outlying envelope=(array([[-1. ],
                                          [-1. ],
                                          [-0.5],
                                          [ 1. ],
                                          [ 1. ],
                                          [ 0.5]]), array([[ 0.5],
                                          [ 0.5],
                                          [ 1. ],
                                          [ 2. ],
                                          [ 1.5],
                                          [ 1. ]])),
            envelopes=[(array([[-1. ],
                               [-1. ],
                               [-0.5],
                               [ 1. ],
                               [ 1. ],
                               [ 0.5]]), array([[ 0.5],
                               [ 0.5],
                               [ 1. ],
                               [ 2. ],
                               [ 1.5],
                               [ 1. ]]))],
            outliers=array([ True, False, False,  True]))

    References:
        .. footbibliography::

    """

    def __init__(
        self,
        fdatagrid: FData,
        chart: Union[Figure, Axes, None] = None,
        *,
        depth_method: Optional[Depth[FDataGrid]] = None,
        prob: Sequence[float] = (0.5,),
        factor: float = 1.5,
        fig: Optional[Figure] = None,
        axes: Optional[Axes] = None,
        n_rows: Optional[int] = None,
        n_cols: Optional[int] = None,
    ):
        """Initialize the Boxplot class.

        Args:
            fdatagrid: Object containing the data.
            depth_method: Method used to order the data.
                Defaults to :func:`modified band depth
                <skfda.exploratory.depth.ModifiedBandDepth>`.
            prob: List with float numbers (in the
                range from 1 to 0) that indicate which central regions to
                represent.
                Defaults to [0.5] which represents the 50% central region.
            factor: Number used to calculate the outlying envelope.
            chart: figure over with the graphs are plotted or axis over
                where the graphs are plotted. If None and ax is also
                None, the figure is initialized.
            fig: figure over with the graphs are
                plotted in case ax is not specified. If None and ax is also
                None, the figure is initialized.
            axes: axis over where the graphs
                are plotted. If None, see param fig.
            n_rows: designates the number of rows of the figure
                to plot the different dimensions of the image. Only specified
                if fig and ax are None.
            n_cols: designates the number of columns of the
                figure to plot the different dimensions of the image. Only
                specified if fig and ax are None.

        """
        super().__init__(
            chart,
            fig=fig,
            axes=axes,
            n_rows=n_rows,
            n_cols=n_cols,
            factor=factor,
        )

        if fdatagrid.dim_domain != 1:
            raise ValueError(
                "Function only supports FDataGrid with domain dimension 1.",
            )

        if sorted(prob, reverse=True) != list(prob):
            raise ValueError(
                "Probabilities required to be in descending order.",
            )

        if min(prob) < 0 or max(prob) > 1:
            raise ValueError("Probabilities must be between 0 and 1.")

        if depth_method is None:
            depth_method = ModifiedBandDepth()
        depth = depth_method(fdatagrid)
        indices_descending_depth = (-depth).argsort(axis=0)

        # The median is the deepest curve
        median_fdata = fdatagrid[indices_descending_depth[0]]
        self._median = median_fdata.data_matrix[0, ...]

        # Central region and envelope must be computed for outlier detection
        central_region = _envelopes.compute_region(
            fdatagrid,
            indices_descending_depth,
            0.5,
        )
        self._central_envelope = _envelopes.compute_envelope(central_region)

        # Non-outlying envelope
        non_outlying_threshold = _envelopes.non_outlying_threshold(
            self._central_envelope,
            factor,
        )
        predicted_outliers = _envelopes.predict_outliers(
            fdatagrid,
            non_outlying_threshold,
        )
        inliers = fdatagrid[predicted_outliers == 0]
        self._non_outlying_envelope = _envelopes.compute_envelope(inliers)

        # Outliers
        self._outliers = _envelopes.predict_outliers(
            fdatagrid,
            self._non_outlying_envelope,
        )

        self._envelopes = [
            _envelopes.compute_envelope(
                _envelopes.compute_region(
                    fdatagrid,
                    indices_descending_depth,
                    p,
                ),
            )
            for p in prob
        ]

        self._fdatagrid = fdatagrid
        self._prob = prob
        self._colormap = plt.cm.get_cmap('RdPu')
        self.barcol = "blue"
        self.outliercol = "red"
        self.mediancol = "black"
        self._show_full_outliers = False

    @property
    def fdatagrid(self) -> FDataGrid:
        return self._fdatagrid

    @property
    def median(self) -> NDArrayFloat:
        return self._median

    @property
    def central_envelope(self) -> Tuple[NDArrayFloat, NDArrayFloat]:
        return self._central_envelope

    @property
    def non_outlying_envelope(self) -> Tuple[NDArrayFloat, NDArrayFloat]:
        return self._non_outlying_envelope

    @property
    def envelopes(self) -> Sequence[Tuple[NDArrayFloat, NDArrayFloat]]:
        return self._envelopes

    @property
    def outliers(self) -> NDArrayBool:
        return self._outliers

    @property
    def show_full_outliers(self) -> bool:
        return self._show_full_outliers

    @show_full_outliers.setter
    def show_full_outliers(self, boolean: bool) -> None:
        if not isinstance(boolean, bool):
            raise ValueError("show_full_outliers must be boolean type")
        self._show_full_outliers = boolean

    @property
    def n_subplots(self) -> int:
        return self.fdatagrid.dim_codomain

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:

        tones = np.linspace(0.1, 1.0, len(self._prob) + 1, endpoint=False)[1:]
        color = self.colormap(tones)

        if self.show_full_outliers:
            var_zorder = 1
        else:
            var_zorder = 4

        outliers = self.fdatagrid[self.outliers]

        grid_points = self.fdatagrid.grid_points[0]

        for m, ax in enumerate(axes):

            # Outliers
            for o in outliers:
                ax.plot(
                    grid_points,
                    o.data_matrix[0, :, m],
                    color=self.outliercol,
                    linestyle='--',
                    zorder=1,
                )

            for envelop, col in zip(self.envelopes, color):
                # central regions
                ax.fill_between(
                    grid_points,
                    envelop[0][..., m],
                    envelop[1][..., m],
                    facecolor=col,
                    zorder=var_zorder,
                )

            # outlying envelope
            ax.plot(
                grid_points,
                self.non_outlying_envelope[0][..., m],
                grid_points,
                self.non_outlying_envelope[1][..., m],
                color=self.barcol,
                zorder=4,
            )

            # central envelope
            ax.plot(
                grid_points,
                self.central_envelope[0][..., m],
                grid_points,
                self.central_envelope[1][..., m],
                color=self.barcol,
                zorder=4,
            )

            # vertical lines
            index = math.ceil(len(grid_points) / 2)
            x = grid_points[index]
            ax.plot(
                [x, x],
                [
                    self.non_outlying_envelope[0][..., m][index],
                    self.central_envelope[0][..., m][index],
                ],
                color=self.barcol,
                zorder=4,
            )
            ax.plot(
                [x, x],
                [
                    self.non_outlying_envelope[1][..., m][index],
                    self.central_envelope[1][..., m][index],
                ],
                color=self.barcol,
                zorder=4,
            )

            # median sample
            ax.plot(
                grid_points,
                self.median[..., m],
                color=self.mediancol,
                zorder=5,
            )

        _set_labels(self.fdatagrid, fig, axes)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return (
            f"Boxplot("
            f"\nFDataGrid={repr(self.fdatagrid)},"
            f"\nmedian={repr(self.median)},"
            f"\ncentral envelope={repr(self.central_envelope)},"
            f"\nnon-outlying envelope={repr(self.non_outlying_envelope)},"
            f"\nenvelopes={repr(self.envelopes)},"
            f"\noutliers={repr(self.outliers)})"
        ).replace('\n', '\n    ')


class SurfaceBoxplot(FDataBoxplot):
    r"""
    Representation of the surface boxplot.

    Class implementing the surface boxplot. Analogously to the functional
    boxplot, it is an informative exploratory tool for visualizing functional
    data with :term:`domain` dimension 2. Nevertheless, it does not implement
    the enhanced surface boxplot.

    Based on the center outward ordering induced by a
    :ref:`depth measure <depth-measures>`
    for functional data, it represents the envelope of the
    50% central region, the median curve, and the maximum non-outlying
    envelope :footcite:`sun+genton_2011_boxplots`.

    Args:
        fdatagrid: Object containing the data.
        method: Method
            used to order the data. Defaults to :class:`modified band depth
            <skfda.exploratory.depth.ModifiedBandDepth>`.
        prob: List with float numbers (in the
            range from 1 to 0) that indicate which central regions to
            represent.
            Defaults to [0.5] which represents the 50% central region.
        factor: Number used to calculate the outlying envelope.

    Attributes:
        fdatagrid: Object containing the data.
        median: contains
            the median/s.
        central_envelope: contains the central envelope/s.
        non_outlying_envelope: contains the non-outlying envelope/s.
        colormap: Colormap from
            which the colors to represent the central regions are selected.
        boxcol: Color of the box, which includes median and central
            envelope.
        outcol: Color of the outlying envelope.
        fig: Figure over with the graphs are
            plotted in case ax is not specified. If None and ax is also
            None, the figure is initialized.
        axes: Axis over where the graphs
            are plotted. If None, see param fig.
        n_rows: Designates the number of rows of the figure
            to plot the different dimensions of the image. Only specified
            if fig and ax are None.
        n_cols: Designates the number of columns of the
            figure to plot the different dimensions of the image. Only
            specified if fig and ax are None.

    Examples:
        Function :math:`f : \mathbb{R^2}\longmapsto\mathbb{R}`.

        >>> from skfda import FDataGrid
        >>> data_matrix = [[[[1], [0.7], [1]],
        ...                 [[4], [0.4], [5]]],
        ...                [[[2], [0.5], [2]],
        ...                 [[3], [0.6], [3]]]]
        >>> grid_points = [[2, 4], [3, 6, 8]]
        >>> fd = FDataGrid(data_matrix, grid_points, dataset_name="dataset",
        ...                argument_names=["x1_label", "x2_label"],
        ...                coordinate_names=["y_label"])
        >>> SurfaceBoxplot(fd)
        SurfaceBoxplot(
            FDataGrid=FDataGrid(
                array([[[[ 1. ],
                         [ 0.7],
                         [ 1. ]],
                        [[ 4. ],
                         [ 0.4],
                         [ 5. ]]],
                       [[[ 2. ],
                         [ 0.5],
                         [ 2. ]],
                        [[ 3. ],
                         [ 0.6],
                         [ 3. ]]]]),
                grid_points=(array([ 2., 4.]), array([ 3., 6., 8.])),
                domain_range=((2.0, 4.0), (3.0, 8.0)),
                dataset_name='dataset',
                argument_names=('x1_label', 'x2_label'),
                coordinate_names=('y_label',),
                extrapolation=None,
                ...),
            median=array([[[ 1. ],
                           [ 0.7],
                           [ 1. ]],
                          [[ 4. ],
                           [ 0.4],
                           [ 5. ]]]),
            central envelope=(array([[[ 1. ],
                                      [ 0.7],
                                      [ 1. ]],
                                     [[ 4. ],
                                      [ 0.4],
                                      [ 5. ]]]),
                              array([[[ 1. ],
                                      [ 0.7],
                                      [ 1. ]],
                                     [[ 4. ],
                                      [ 0.4],
                                      [ 5. ]]])),
            outlying envelope=(array([[[ 1. ],
                                       [ 0.7],
                                       [ 1. ]],
                                      [[ 4. ],
                                       [ 0.4],
                                       [ 5. ]]]),
                               array([[[ 1. ],
                                       [ 0.7],
                                       [ 1. ]],
                                      [[ 4. ],
                                       [ 0.4],
                                       [ 5. ]]])))

    References:
        .. footbibliography::

    """

    def __init__(
        self,
        fdatagrid: FDataGrid,
        chart: Union[Figure, Axes, None] = None,
        *,
        depth_method: Optional[Depth[FDataGrid]] = None,
        factor: float = 1.5,
        fig: Optional[Figure] = None,
        axes: Optional[Axes] = None,
        n_rows: Optional[int] = None,
        n_cols: Optional[int] = None,
    ) -> None:

        super().__init__(
            chart,
            fig=fig,
            axes=axes,
            n_rows=n_rows,
            n_cols=n_cols,
            factor=factor,
        )

        if fdatagrid.dim_domain != 2:
            raise ValueError(
                "Class only supports FDataGrid with domain dimension 2.",
            )

        if depth_method is None:
            depth_method = ModifiedBandDepth()

        depth = depth_method(fdatagrid)
        indices_descending_depth = (-depth).argsort(axis=0)

        # The mean is the deepest curve
        self._median = fdatagrid.data_matrix[indices_descending_depth[0]]

        # Central region and envelope must be computed for outlier detection
        central_region = _envelopes.compute_region(
            fdatagrid,
            indices_descending_depth,
            0.5,
        )
        self._central_envelope = _envelopes.compute_envelope(central_region)

        # Non-outlying envelope
        non_outlying_threshold = _envelopes.non_outlying_threshold(
            self._central_envelope,
            factor,
        )
        predicted_outliers = _envelopes.predict_outliers(
            fdatagrid,
            non_outlying_threshold,
        )
        inliers = fdatagrid[predicted_outliers == 0]
        self._non_outlying_envelope = _envelopes.compute_envelope(inliers)

        self._fdatagrid = fdatagrid
        self.colormap = plt.cm.get_cmap('Greys')
        self._boxcol = 1.0
        self._outcol = 0.7

    @property
    def fdatagrid(self) -> FDataGrid:
        return self._fdatagrid

    @property
    def median(self) -> NDArrayFloat:
        return self._median

    @property
    def central_envelope(self) -> Tuple[NDArrayFloat, NDArrayFloat]:
        return self._central_envelope

    @property
    def non_outlying_envelope(self) -> Tuple[NDArrayFloat, NDArrayFloat]:
        return self._non_outlying_envelope

    @property
    def boxcol(self) -> float:
        return self._boxcol

    @boxcol.setter
    def boxcol(self, value: float) -> None:
        if value < 0 or value > 1:
            raise ValueError("boxcol must be a number between 0 and 1.")

        self._boxcol = value

    @property
    def outcol(self) -> float:
        return self._outcol

    @outcol.setter
    def outcol(self, value: float) -> None:
        if value < 0 or value > 1:
            raise ValueError("outcol must be a number between 0 and 1.")
        self._outcol = value

    @property
    def dim(self) -> int:
        return 3

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:

        x = self.fdatagrid.grid_points[0]
        lx = len(x)
        y = self.fdatagrid.grid_points[1]
        ly = len(y)
        X, Y = np.meshgrid(x, y)

        for m, ax in enumerate(axes):

            # mean sample
            ax.plot_wireframe(
                X,
                Y,
                np.squeeze(self.median[..., m]).T,
                rstride=ly,
                cstride=lx,
                color=self.colormap(self.boxcol),
            )
            ax.plot_surface(
                X,
                Y,
                np.squeeze(self.median[..., m]).T,
                color=self.colormap(self.boxcol),
                alpha=0.8,
            )

            # central envelope
            ax.plot_surface(
                X,
                Y,
                np.squeeze(self.central_envelope[0][..., m]).T,
                color=self.colormap(self.boxcol),
                alpha=0.5,
            )
            ax.plot_wireframe(
                X,
                Y,
                np.squeeze(self.central_envelope[0][..., m]).T,
                rstride=ly,
                cstride=lx,
                color=self.colormap(self.boxcol),
            )
            ax.plot_surface(
                X,
                Y,
                np.squeeze(self.central_envelope[1][..., m]).T,
                color=self.colormap(self.boxcol),
                alpha=0.5,
            )
            ax.plot_wireframe(
                X,
                Y,
                np.squeeze(self.central_envelope[1][..., m]).T,
                rstride=ly,
                cstride=lx,
                color=self.colormap(self.boxcol),
            )

            # box vertical lines
            for indices in (
                (0, 0),
                (0, ly - 1),
                (lx - 1, 0),
                (lx - 1, ly - 1),
            ):
                x_corner = x[indices[0]]
                y_corner = y[indices[1]]
                ax.plot(
                    [x_corner, x_corner],
                    [y_corner, y_corner],
                    [
                        self.central_envelope[1][..., m][
                            indices[0],
                            indices[1],
                        ],
                        self.central_envelope[0][..., m][
                            indices[0],
                            indices[1],
                        ],
                    ],
                    color=self.colormap(self.boxcol),
                )

            # outlying envelope
            ax.plot_surface(
                X,
                Y,
                np.squeeze(self.non_outlying_envelope[0][..., m]).T,
                color=self.colormap(self.outcol),
                alpha=0.3,
            )
            ax.plot_wireframe(
                X,
                Y,
                np.squeeze(self.non_outlying_envelope[0][..., m]).T,
                rstride=ly,
                cstride=lx,
                color=self.colormap(self.outcol),
            )
            ax.plot_surface(
                X,
                Y,
                np.squeeze(self.non_outlying_envelope[1][..., m]).T,
                color=self.colormap(self.outcol),
                alpha=0.3,
            )
            ax.plot_wireframe(
                X,
                Y,
                np.squeeze(self.non_outlying_envelope[1][..., m]).T,
                rstride=ly,
                cstride=lx,
                color=self.colormap(self.outcol),
            )

            # vertical lines from central to outlying envelope
            x_index = math.floor(lx / 2)
            x_central = x[x_index]
            y_index = math.floor(ly / 2)
            y_central = y[y_index]
            ax.plot(
                [x_central, x_central],
                [y_central, y_central],
                [
                    self.non_outlying_envelope[1][..., m][x_index, y_index],
                    self.central_envelope[1][..., m][x_index, y_index],
                ],
                color=self.colormap(self.boxcol),
            )
            ax.plot(
                [x_central, x_central],
                [y_central, y_central],
                [
                    self.non_outlying_envelope[0][..., m][x_index, y_index],
                    self.central_envelope[0][..., m][x_index, y_index],
                ],
                color=self.colormap(self.boxcol),
            )

        _set_labels(self.fdatagrid, fig, axes)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return (
            f"SurfaceBoxplot("
            f"\nFDataGrid={repr(self.fdatagrid)},"
            f"\nmedian={repr(self.median)},"
            f"\ncentral envelope={repr(self.central_envelope)},"
            f"\noutlying envelope={repr(self.non_outlying_envelope)})"
        ).replace('\n', '\n    ')



"""Registration of functional data module.

This module contains methods to perform the registration of
functional data, in basis as well in discretized form.
"""

from ..._utils import invert_warping, normalize_warping
from . import validation
from ._fisher_rao import ElasticRegistration, FisherRaoElasticRegistration
from ._landmark_registration import (
    landmark_elastic_registration,
    landmark_elastic_registration_warping,
    landmark_registration,
    landmark_shift,
    landmark_shift_deltas,
    landmark_shift_registration,
)
from ._lstsq_shift_registration import (
    LeastSquaresShiftRegistration,
    ShiftRegistration,
)



"""Test the math module."""
import unittest
from typing import Sequence

import numpy as np

import skfda
from skfda._utils import _pairwise_symmetric
from skfda.representation._typing import NDArrayFloat
from skfda.representation.basis import Monomial, Tensor, VectorValued


def _ndm(*args: NDArrayFloat) -> Sequence[NDArrayFloat]:
    return [
        x[(None,) * i + (slice(None),) + (None,) * (len(args) - i - 1)]
        for i, x in enumerate(args)
    ]


class InnerProductTest(unittest.TestCase):
    """Tests for the inner product."""

    def test_several_variables(self) -> None:
        """Test inner_product with functions of several variables."""
        def f(
            x: NDArrayFloat,
            y: NDArrayFloat,
            z: NDArrayFloat,
        ) -> NDArrayFloat:
            return x * y * z

        t = np.linspace(0, 1, 30)

        x2, y2, z2 = _ndm(t, 2 * t, 3 * t)

        data_matrix = f(x2, y2, z2)

        grid_points = [t, 2 * t, 3 * t]

        fd = skfda.FDataGrid(
            data_matrix[np.newaxis, ...],
            grid_points=grid_points,
        )

        basis = Tensor([
            Monomial(n_basis=5, domain_range=(0, 1)),
            Monomial(n_basis=5, domain_range=(0, 2)),
            Monomial(n_basis=5, domain_range=(0, 3)),
        ])

        fd_basis = fd.to_basis(basis)

        res = 8

        np.testing.assert_allclose(
            skfda.misc.inner_product(fd, fd),
            res,
            rtol=1e-4,
        )
        np.testing.assert_allclose(
            skfda.misc.inner_product(fd_basis, fd_basis),
            res,
            rtol=1e-4,
        )

    def test_vector_valued(self) -> None:
        """Test inner_product with vector valued functions."""
        def f(x: NDArrayFloat) -> NDArrayFloat:
            return x**2

        def g(y: NDArrayFloat) -> NDArrayFloat:
            return 3 * y

        t = np.linspace(0, 1, 100)

        data_matrix = np.array([np.array([f(t), g(t)]).T])

        grid_points = [t]

        fd = skfda.FDataGrid(
            data_matrix,
            grid_points=grid_points,
        )

        basis = VectorValued([
            Monomial(n_basis=5),
            Monomial(n_basis=5),
        ])

        fd_basis = fd.to_basis(basis)

        res = 1 / 5 + 3

        np.testing.assert_allclose(
            skfda.misc.inner_product(fd, fd),
            res,
            rtol=1e-5,
        )
        np.testing.assert_allclose(
            skfda.misc.inner_product(fd_basis, fd_basis),
            res,
            rtol=1e-5,
        )

    def test_matrix(self) -> None:
        """Test inner_product_matrix function."""
        basis = skfda.representation.basis.BSpline(n_basis=12)

        X = skfda.datasets.make_gaussian_process(
            n_samples=10,
            n_features=20,
            cov=skfda.misc.covariances.Gaussian(),
            random_state=0,
        )
        Y = skfda.datasets.make_gaussian_process(
            n_samples=10,
            n_features=20,
            cov=skfda.misc.covariances.Gaussian(),
            random_state=1,
        )

        X_basis = X.to_basis(basis)
        Y_basis = Y.to_basis(basis)

        gram = skfda.misc.inner_product_matrix(X, Y)
        gram_basis = skfda.misc.inner_product_matrix(X_basis, Y_basis)

        np.testing.assert_allclose(gram, gram_basis, rtol=1e-2)

        gram_pairwise = _pairwise_symmetric(
            skfda.misc.inner_product, X, Y)

        np.testing.assert_allclose(gram, gram_pairwise)


class CosineSimilarityVectorTest(unittest.TestCase):
    """Tests for cosine similarity for vectors."""

    def setUp(self) -> None:
        self.arr = np.array([
            [0, 0, 1],
            [1, 1, 1],
            [1, 2, 3],
            [1, 0, 1],
        ])

        self.arr_samelen = np.array([
            [2, 4, 1],
            [7, 2, 1],
            [0, 1, 0],
            [3, 2, 0],
        ])

        self.arr_short = np.array([
            [2, 4, 6],
            [5, 1, 7],
        ])

    def test_cosine_similarity_elementwise(self) -> None:
        """Elementwise example for vectors."""
        np.testing.assert_allclose(
            skfda.misc.inner_product(self.arr, self.arr_samelen),
            [1, 10, 2, 3],
        )

        np.testing.assert_allclose(
            skfda.misc.cosine_similarity(self.arr, self.arr_samelen),
            [
                1 / np.sqrt(21),
                10 / np.sqrt(3) / np.sqrt(54),
                2 / np.sqrt(14),
                3 / np.sqrt(2) / np.sqrt(13),
            ],
        )

    def test_cosine_similarity_matrix_one(self) -> None:
        """Matrix example for vectors with one input."""

        for arr2 in (None, self.arr):

            np.testing.assert_allclose(
                skfda.misc.inner_product_matrix(self.arr, arr2),
                [
                    [1, 1, 3, 1],
                    [1, 3, 6, 2],
                    [3, 6, 14, 4],
                    [1, 2, 4, 2],
                ],
            )

            np.testing.assert_allclose(
                skfda.misc.cosine_similarity_matrix(self.arr, arr2),
                [
                    [
                        1,
                        1 / np.sqrt(3),
                        3 / np.sqrt(14),
                        1 / np.sqrt(2),
                    ],
                    [
                        1 / np.sqrt(3),
                        3 / np.sqrt(3 * 3),
                        6 / np.sqrt(3 * 14),
                        2 / np.sqrt(3 * 2),
                    ],
                    [
                        3 / np.sqrt(14),
                        6 / np.sqrt(14 * 3),
                        14 / np.sqrt(14 * 14),
                        4 / np.sqrt(14 * 2),
                    ],
                    [
                        1 / np.sqrt(2),
                        2 / np.sqrt(2 * 3),
                        4 / np.sqrt(2 * 14),
                        2 / np.sqrt(2 * 2),
                    ],
                ],
            )

    def test_cosine_similarity_matrix_two(self) -> None:
        """Matrix example for vectors with two inputs."""
        np.testing.assert_allclose(
            skfda.misc.inner_product_matrix(self.arr, self.arr_short),
            [
                [6, 7],
                [12, 13],
                [28, 28],
                [8, 12],
            ],
        )

        np.testing.assert_allclose(
            skfda.misc.cosine_similarity_matrix(self.arr, self.arr_short),
            [
                [
                    6 / np.sqrt(56),
                    7 / np.sqrt(75),
                ],
                [
                    12 / np.sqrt(3) / np.sqrt(56),
                    13 / np.sqrt(3) / np.sqrt(75),
                ],
                [
                    28 / np.sqrt(14) / np.sqrt(56),
                    28 / np.sqrt(14) / np.sqrt(75),
                ],
                [
                    8 / np.sqrt(2) / np.sqrt(56),
                    12 / np.sqrt(2) / np.sqrt(75),
                ],
            ],
        )


if __name__ == "__main__":
    unittest.main()
